import logging

from aiohttp import WSMsgType

from uxapi import UXTopic
from uxapi import Session
from uxapi import Awaitables
from uxapi import listiter


class WSHandler:
    logger = logging.getLogger(__name__)

    def __init__(self, exchange, wsurl, topic_set):
        self.exchange = exchange
        self.wsurl = wsurl
        self.topic_set = topic_set
        self.session = None
        self.own_session = False
        self.ws = None
        self.pending_topics = None
        self.awaitables = Awaitables()
        self.pre_processors = listiter([])

    def get_credentials(self):
        credentials = self.exchange.requiredCredentials
        requires = [key for key in credentials if credentials[key]]
        result = {}
        for key in requires:
            value = getattr(self.exchange, key)
            if value:
                result[key] = value
            else:
                raise RuntimeError(f'requires `{key}`')
        return result

    async def run(self, collector=None):
        try:
            self.ws = await self.connect()
            self.prepare()
            await self.do_run(collector)
        finally:
            await self.cleanup()

    async def do_run(self, collector):
        while True:
            if 'recv' not in self.awaitables:
                self.awaitables.create_task(self.recv(), 'recv')
            name, result = await self.awaitables.wait()
            if name == 'recv' and result is not None:
                try:
                    msg = self.pre_process(result)
                except StopIteration:
                    continue
                else:
                    if collector:
                        collector(msg)

    async def connect(self):
        if not self.session:
            self.session = Session()
            self.own_session = True
        ws = await self.session.ws_connect(self.wsurl)
        self.on_connected()
        return ws

    def prepare(self):
        self.create_keepalive_task()
        if self.login_required:
            self.create_login_task()
        else:
            self.on_prepared()

    def pre_process(self, data):
        msg = self.decode(data)
        self.pre_processors.rewind()
        for processor in self.pre_processors:
            msg = processor(msg)
        return msg

    def on_connected(self):
        pass

    def create_keepalive_task(self):
        self.pre_processors.prepend(self.on_keepalive_message)
        return self.awaitables.create_task(self.keepalive(), 'keepalive')

    async def keepalive(self):
        raise NotImplementedError

    def on_keepalive_message(self, message):
        raise NotImplementedError

    @property
    def login_required(self):
        return False

    def create_login_task(self):
        self.pre_processors.append(self.on_login_message)
        credentials = self.get_credentials()
        return self.awaitables.create_task(self.login(credentials), 'login')

    async def login(self, credentials):
        command = self.login_command(credentials)
        await self.send(command)

    def login_command(self, credentials):
        raise NotImplementedError

    def on_login_message(self, message):
        raise NotImplementedError

    def on_logged_in(self):
        self.pre_processors.remove()
        self.on_prepared()

    def on_prepared(self):
        self.create_subscribe_task()

    def create_subscribe_task(self):
        self.pre_processors.append(self.on_subscribe_message)
        topic_set = {self.convert_topic(topic) for topic in self.topic_set}
        self.pending_topics = topic_set
        return self.awaitables.create_task(self.subscribe(topic_set), 'subscribe')

    def convert_topic(self, topic: UXTopic):
        return self.exchange.convert_topic(topic)

    async def subscribe(self, topic_set):
        commands = self.subscribe_commands(topic_set)
        for command in commands:
            await self.send(command)

    def subscribe_commands(self, topic_set):
        raise NotImplementedError

    def on_subscribe_message(self, message):
        raise NotImplementedError

    def on_subscribed(self, topic):
        self.pending_topics.remove(topic)
        if not self.pending_topics:
            self.pre_processors.remove()
            self.pending_topics = None

    async def send(self, command):
        if isinstance(command, dict):
            await self.ws.send_json(command)
        else:
            await self.ws.send_str(command)

    async def recv(self):
        wsmsg = await self.ws.receive()
        if wsmsg.type in (WSMsgType.BINARY, WSMsgType.TEXT):
            return wsmsg.data
        else:
            raise RuntimeError(f'unexpected message: {wsmsg}')

    def decode(self, data):
        return data

    async def cleanup(self):
        await self.awaitables.cleanup()

        if self.ws:
            try:
                await self.ws.close()
            except Exception:
                pass

        if self.session and self.own_session:
            try:
                await self.session.close()
            except Exception:
                pass



import json
import zlib
import time
import asyncio
import bisect
from itertools import zip_longest, chain
import binascii

import pendulum

from uxapi import register_exchange
from uxapi.exchanges.ccxt.okex import okex
from uxapi import UXSymbol
from uxapi import WSHandler
from uxapi import UXPatch
from uxapi.helpers import (
    hmac,
    deep_extend,
    contract_delivery_time,
)


@register_exchange('okex')
class Okex(UXPatch, okex):
    def __init__(self, market_type, config=None):
        fetch_market_option = ['spot'] if market_type == 'margin' else [market_type]
        return super().__init__(market_type, deep_extend({
            'options': {
                'fetchMarkets': fetch_market_option,
                'defaultType': market_type,
            }
        }, config or {}))

    def describe(self):
        return self.deep_extend(super().describe(), {
            'deliveryHourUTC': 8,

            'urls': {
                'wsapi': 'wss://real.okex.com:8443/ws/v3',
            },

            'wsapi': {
                'spot': {
                    'ticker': 'spot/ticker:{symbol}',
                    'orderbook': 'spot/depth{level}:{symbol}',
                    'ohlcv': 'spot/candle{period_in_sec}s:{symbol}',
                    'trade': 'spot/trade:{symbol}',
                    'myorder': 'spot/order:{symbol}',
                    'account': 'spot/account:{currency}',
                    'margin_account': 'spot/margin_account:{symbol}',
                },
                'futures': {
                    'ticker': 'futures/ticker:{symbol}',
                    'orderbook': 'futures/depth{level}:{symbol}',
                    'ohlcv': 'futures/candle{period_in_sec}s:{symbol}',
                    'trade': 'futures/trade:{symbol}',
                    'myorder': 'futures/order:{symbol}',
                    'position': 'futures/position:{symbol}',
                    'account': 'futures/account:{currency}',
                    'estimated_price': 'futures/estimated_price:{symbol}',
                    'instruments': 'futures/instruments',
                    'price_range': 'futures/price_range:{symbol}',
                    'mark_price': 'futures/mark_price:{symbol}',
                },
                'swap': {
                    'ticker': 'swap/ticker:{symbol}',
                    'orderbook': 'swap/depth{level}:{symbol}',
                    'ohlcv': 'swap/candle{period_in_sec}s:{symbol}',
                    'trade': 'swap/trade:{symbol}',
                    'myorder': 'swap/order:{symbol}',
                    'position': 'swap/position:{symbol}',
                    'account': 'swap/account:{symbol}',
                    'order_algo': 'swap/order_algo:{symbol}',
                    'funding_rate': 'swap/funding_rate:{symbol}',
                    'price_range': 'swap/price_range:{symbol}',
                    'mark_price': 'swap/mark_price:{symbol}',
                },
                'option': {
                    'position': 'option/position:{currency}-USD',
                    'account': 'option/account:{currency}-USD',
                    'myorder': 'option/order:{currency}-USD',
                    'instruments': 'option/instruments:{currency}-USD',
                    'summary': 'option/summary:{currency}-USD',
                    'ticker': 'option/ticker:{symbol}',
                    'orderbook': 'option/depth{level}:{symbol}',
                    'trade': 'option/trade:{symbol}',
                    'ohlcv': 'option/candle{period_in_sec}s:{symbol}',
                },
                'index': {
                    'ticker': 'index/ticker:{symbol}',
                    'ohlcv': 'index/candle{period_in_sec}s:{symbol}',
                }
            }
        })

    def _fetch_markets(self, params=None):
        markets = super()._fetch_markets(params)
        for market in markets:
            underlying = self.safe_string(market['info'], 'underlying')
            if underlying and underlying.endswith('-USDT'):
                quote, base = underlying.split('-')
                market['base'] = market['baseId'] = base
                market['quote'] = market['quoteId'] = quote

            contract_value = self.safe_float(market['info'], 'contract_val')
            if contract_value:
                market['contractValue'] = contract_value

            if market['futures']:
                delivery_time = pendulum.parse(self.safe_string(market['info'], 'delivery'))
                delivery_time = delivery_time.add(hours=self.deliveryHourUTC)
                market['deliveryTime'] = delivery_time.to_iso8601_string()
            elif market['option']:
                market['deliveryTime'] = market['info']['delivery']
        return markets

    def order_book_merger(self):
        return OkexOrderBookMerger()

    def wshandler(self, topic_set):
        return OkexWSHandler(self, self.urls['wsapi'], topic_set)

    def convert_symbol(self, uxsymbol):
        if uxsymbol.market_type in ['spot', 'margin']:
            return uxsymbol.name

        if uxsymbol.market_type == 'swap':
            if uxsymbol.base == 'USDT':
                return f'{uxsymbol.quote}-{uxsymbol.base}-SWAP'
            elif uxsymbol.quote == 'USD':
                return f'{uxsymbol.base}-{uxsymbol.quote}-SWAP'

        if uxsymbol.market_type == 'futures':
            delivery_time = contract_delivery_time(
                expiration=uxsymbol.contract_expiration,
                delivery_hour=self.deliveryHourUTC)
            if uxsymbol.base == 'USDT':
                return f'{uxsymbol.quote}-{uxsymbol.base}-{delivery_time:%y%m%d}'
            elif uxsymbol.quote == 'USD':
                return f'{uxsymbol.base}-{uxsymbol.quote}-{delivery_time:%y%m%d}'

        if uxsymbol.market_type == 'index':
            return f'{uxsymbol.base}-{uxsymbol.quote}'

        if uxsymbol.market_type == 'option':
            base = uxsymbol.base
            dt = pendulum.parse(uxsymbol.contract_expiration)
            expiration = dt.format('YYMMDD')
            strike_price = uxsymbol.option_strike_price
            call_or_put = uxsymbol.option_type
            return f'{base}-USD-{expiration}-{strike_price}-{call_or_put}'

        raise ValueError(f'invalid symbol: {uxsymbol}')

    def convert_topic(self, uxtopic):
        maintype = uxtopic.maintype
        subtypes = uxtopic.subtypes
        template = self.wsapi[self.market_type][maintype]

        if maintype == 'instruments':
            return template

        params = {}
        if 'currency' in template:
            params['currency'] = uxtopic.extrainfo
        else:
            exchange_id, market_type, _, extrainfo = uxtopic
            uxsymbol = UXSymbol(exchange_id, market_type, extrainfo)
            params['symbol'] = self.market_id(uxsymbol)

        if maintype == 'ohlcv':
            params['period_in_sec'] = self.timeframes[subtypes[0]]

        if maintype == 'orderbook':
            if not subtypes:
                params['level'] = '5'
            elif subtypes[0] in ('full', 'tbt'):
                params['level'] = '_l2_tbt'
            else:
                params['level'] = ''
        return template.format(**params)


class OkexWSHandler(WSHandler):
    def on_connected(self):
        self.pre_processors.append(self.on_error_message)

    def on_error_message(self, msg):
        if msg.get('event') == 'error':
            err_msg = msg['message']
            err_code = msg['errorCode']
            raise RuntimeError(f'{err_msg}({err_code})')
        else:
            return msg

    def create_keepalive_task(self):
        self.last_message_timestamp = time.time()
        return super().create_keepalive_task()

    async def keepalive(self):
        interval = 10
        while True:
            await asyncio.sleep(interval)
            now = time.time()
            if now - self.last_message_timestamp >= interval:
                await self.send('ping')

    def on_keepalive_message(self, msg):
        self.last_message_timestamp = time.time()
        if msg == 'pong':
            raise StopIteration
        else:
            return msg

    @property
    def login_required(self):
        private_types = {'myorder', 'position', 'account', 'margin_account'}
        topic_types = {topic.maintype for topic in self.topic_set}
        for type in private_types:
            if type in topic_types:
                return True
        return False

    def login_command(self, credentials):
        server_timestamp = time.time()
        payload = f'{server_timestamp}GET/users/self/verify'
        signature = hmac(
            bytes(credentials['secret'], 'utf8'),
            bytes(payload, 'utf8'),
        )
        return {
            'op': 'login',
            'args': [
                credentials['apiKey'],
                credentials['password'],
                server_timestamp,
                signature.decode(),
            ]
        }

    def on_login_message(self, msg):
        if msg.get('event') == 'login':
            self.logger.info(f'logged in')
            self.on_logged_in()
            raise StopIteration
        else:
            return msg

    def subscribe_commands(self, topic_set):
        command = {
            'op': 'subscribe',
            'args': list(topic_set),
        }
        return [command]

    def on_subscribe_message(self, msg):
        if msg.get('event') == 'subscribe':
            topic = msg['channel']
            self.logger.info(f'{topic} subscribed')
            self.on_subscribed(topic)
            raise StopIteration
        else:
            return msg

    def decode(self, data):
        bytes_ = zlib.decompress(data, wbits=-zlib.MAX_WBITS)
        msg = bytes_.decode()
        try:
            jsonmsg = json.loads(msg)
        except json.JSONDecodeError:
            return msg
        else:
            return jsonmsg


class OkexOrderBookMerger:
    def __init__(self):
        self.snapshot = None
        self.prices = None

    def __call__(self, patch):
        if patch['action'] == 'partial':
            self.on_snapshot(patch)
        elif patch['action'] == 'update':
            self.merge(patch)
        else:
            raise ValueError('unexpected action')
        self.validate()
        return self.snapshot

    def on_snapshot(self, snapshot):
        data = snapshot['data'][-1]
        self.snapshot = snapshot
        self.snapshot['data'] = [data]
        self.prices = {
            'asks': [float(item[0]) for item in data['asks']],
            'bids': [-float(item[0]) for item in data['bids']]
        }

    def merge(self, patch):
        snapshot_data = self.snapshot['data'][0]
        patch_data_list = patch['data']
        for patch_data in patch_data_list:
            snapshot_data['timestamp'] = patch_data['timestamp']
            snapshot_data['checksum'] = patch_data['checksum']
            self.merge_asks_bids(snapshot_data['asks'], patch_data['asks'],
                                 self.prices['asks'], False)
            self.merge_asks_bids(snapshot_data['bids'], patch_data['bids'],
                                 self.prices['bids'], True)

    def merge_asks_bids(self, snapshot_lst, patch_lst, price_lst, negative_price):
        for item in patch_lst:
            price, amount = float(item[0]), float(item[1])
            if negative_price:
                price = -price
            i = bisect.bisect_left(price_lst, price)
            if i != len(price_lst) and price_lst[i] == price:
                if amount == 0:
                    price_lst.pop(i)
                    snapshot_lst.pop(i)
                else:
                    snapshot_lst[i] = item
            else:
                if amount != 0:
                    price_lst.insert(i, price)
                    snapshot_lst.insert(i, item)

    def validate(self):
        data = self.snapshot['data'][0]
        asks = (item[:2] for item in data['asks'][:25])
        bids = (item[:2] for item in data['bids'][:25])
        items = filter(None, chain(*zip_longest(bids, asks)))
        text = ':'.join(chain(*items))
        crc32 = binascii.crc32(text.encode())
        checksum = data['checksum'] & 0xffffffff
        if crc32 != checksum:
            raise RuntimeError('invalid order book data')



import time
import asyncio
import json
import operator

import ccxt

from uxapi import register_exchange
from uxapi import UXSymbol
from uxapi import UXPatch
from uxapi import WSHandler
from uxapi.helpers import (
    hmac,
    contract_delivery_time
)


@register_exchange('bitmex')
class Bitmex(UXPatch, ccxt.bitmex):
    def describe(self):
        return self.deep_extend(super().describe(), {
            'deliveryHourUTC': 12,

            'options': {
                'ws-api-expires': 60*60*24*1000,  # 1000 days
            },

            'urls': {
                'wsapi': 'wss://www.bitmex.com/realtime',
                'wsapiTest': 'wss://testnet.bitmex.com/realtime',
            },

            'wsapi': [
                'orderbook',
                'trade',
                'quote',
                'announcement',
                'chat',
                'connected',
                'funding',
                'instrument',
                'insurance',
                'liquidation',
                'publicNotifications',
                'settlement',
                'affiliate',
                'execution',
                'myorder',
                'margin',
                'position',
                'privateNotifications',
                'transact',
                'wallet',
            ],
        })

    @classmethod
    def testnet(cls, market_type, config):
        bitmex = cls(market_type, config)
        bitmex.urls['api'] = bitmex.urls['test']
        bitmex.urls['wsapi'] = bitmex.urls['wsapiTest']
        return bitmex

    def _fetch_markets(self, params):
        markets = super()._fetch_markets(params)
        for market in markets:
            contract_value = self.safe_float(market['info'], 'lotSize')
            if contract_value:
                market['contractValue'] = contract_value
            if market['type'] == 'future':
                market['type'] == 'futures'
                market['deliveryTime'] = market['info']['settle']
        return markets

    def parse_order_book(self, orderbook):
        # [
        #     {
        #         "symbol": "XBTUSD",
        #         "id": 15599187800,
        #         "side": "Sell",
        #         "size": 542,
        #         "price": 8122
        #     },
        #     ...
        # ]
        asks = []
        bids = []
        for item in orderbook:
            price = float(item['price'])
            amount = float(item['size'])
            if item['side'] == 'Sell':
                asks.append([price, amount])
            else:
                bids.append([price, amount])
        asks = sorted(asks, key=operator.itemgetter(0))
        bids = sorted(bids, key=operator.itemgetter(0), reverse=True)
        return {
            'asks': asks,
            'bids': bids,
            'timestamp': None,
            'datetime': None,
            'nonce': None
        }

    def order_book_merger(self):
        return BitmexOrderBookMerger()

    def wshandler(self, topic_set):
        return BitmexWSHandler(self, self.urls['wsapi'], topic_set)

    def convert_symbol(self, uxsymbol):
        if uxsymbol.market_type == 'swap':
            if uxsymbol.name == '!ETHUSD/BTC':
                return 'ETH/USD'
            elif uxsymbol.name == '!XRPUSD/BTC':
                return 'XRP/USD'
            elif uxsymbol.name == 'BTC/USD':
                return uxsymbol.name

        elif uxsymbol.market_type == 'futures':
            if uxsymbol.base == 'BTC':
                delivery_time = contract_delivery_time(
                    expiration=uxsymbol.contract_expiration,
                    delivery_hour=self.deliveryHourUTC)
                code = self._contract_code(delivery_time)
                if uxsymbol.quote == 'USD':
                    return f'XBT{code}'
                elif uxsymbol.quote in ('ADA', 'BCH', 'EOS', 'ETH', 'LTC', 'TRX', 'XRP'):
                    return f'{uxsymbol.quote}{code}'

        elif uxsymbol.market_type == 'index':
            return uxsymbol.name

        raise ValueError('invalid symbol')

    @staticmethod
    def _contract_code(delivery_time):
        month_codes = "FGHJKMNQUVXZ"
        year = delivery_time.year - 2000
        quarter = (delivery_time.month + 2) // 3
        delivery_month = month_codes[quarter * 3 - 1]
        return f'{delivery_month}{year}'

    def convert_topic(self, uxtopic):
        maintype = uxtopic.maintype
        assert maintype in self.wsapi
        subtypes = uxtopic.subtypes
        topic = None
        if maintype == 'orderbook':
            if not subtypes:
                level = '10'
            elif subtypes[0] == 'full':
                level = 'L2'
            elif subtypes[0] == '25':
                level = 'L2_25'
            else:
                raise ValueError('invalid topic')
            topic = f'orderBook{level}'
        elif maintype == 'quote':
            if subtypes:
                topic = f'quoteBin{subtypes[0]}'
            else:
                topic = 'quote'
        elif maintype == 'trade':
            if subtypes:
                topic = f'tradeBin{subtypes[0]}'
            else:
                topic = 'trade'
        elif maintype == 'myorder':
            topic = 'order'
        else:
            topic = maintype

        if uxtopic.extrainfo:
            uxsymbol = UXSymbol(uxtopic.exchange_id, uxtopic.market_type,
                                uxtopic.extrainfo)
            symbol = self.market_id(uxsymbol)
            topic = f'{topic}:{symbol}'

        return topic


class BitmexWSHandler(WSHandler):
    default_api_expires = 60 * 60 * 24 * 1000  # 1000 days

    def on_connected(self):
        self.pre_processors.append(self.on_info_message)
        self.pre_processors.append(self.on_error_message)

    def on_info_message(self, msg):
        if 'info' in msg:
            self.logger.info(msg)
            raise StopIteration
        else:
            return msg

    def on_error_message(self, msg):
        if 'error' in msg:
            raise RuntimeError(msg)
        else:
            return msg

    def create_keepalive_task(self):
        self.last_message_timestamp = time.time()
        return super().create_keepalive_task()

    async def keepalive(self):
        interval = 5
        while True:
            await asyncio.sleep(interval)
            now = time.time()
            if now - self.last_message_timestamp >= interval:
                await self.send('ping')

    def on_keepalive_message(self, msg):
        self.last_message_timestamp = time.time()
        if msg == 'pong':
            raise StopIteration
        else:
            return msg

    @property
    def login_required(self):
        private_types = {
            'myorder', 'margin', 'position', 'affiliate', 'execution',
            'privateNotifications', 'transact', 'wallet',
        }
        topic_types = {topic.maintype for topic in self.topic_set}
        for type in private_types:
            if type in topic_types:
                return True
        return False

    def login_command(self, credentials):
        if self.exchange:
            expires = self.exchange.options['ws-api-expires']
        else:
            expires = self.default_api_expires
        expires += int(time.time())

        payload = 'GET' + '/realtime' + str(expires)
        signature = hmac(
            bytes(credentials['secret'], 'utf8'),
            bytes(payload, 'utf8'),
            digest='hex'
        )
        return {
            'op': 'authKeyExpires',
            'args': [
                credentials['apiKey'],
                expires,
                signature,
            ]
        }

    def on_login_message(self, msg):
        if ('request' in msg) and (msg['request'].get('op') == 'authKeyExpires'):
            if msg['success']:
                self.logger.info('logged in')
                self.on_logged_in()
                raise StopIteration
            else:
                raise RuntimeError('login failed')
        else:
            return msg

    def subscribe_commands(self, topic_set):
        command = {
            'op': 'subscribe',
            'args': list(topic_set),
        }
        return [command]

    def on_subscribe_message(self, msg):
        if 'subscribe' in msg:
            topic = msg['subscribe']
            self.logger.info(f'{topic} subscribed')
            self.on_subscribed(topic)
            raise StopIteration
        else:
            return msg

    def decode(self, data):
        try:
            jsonmsg = json.loads(data)
            return jsonmsg
        except json.JSONDecodeError:
            return data


class BitmexOrderBookMerger:
    def __init__(self):
        self.snapshot = None
        self.data = None

    def __call__(self, patch):
        if patch['action'] == 'partial':
            self.on_snapshot(patch)
        elif patch['action'] in ('update', 'delete', 'insert'):
            self.merge(patch)
        else:
            raise ValueError('unexpected action')
        return self.snapshot

    def on_snapshot(self, snapshot):
        self.snapshot = snapshot
        self.data = {item['id']: item for item in snapshot['data']}
        self.update_snapshot()

    def merge(self, patch):
        if not self.snapshot:
            raise StopIteration

        if patch['action'] == 'update':
            for item in patch['data']:
                self.data[item['id']]['size'] = item['size']

        elif patch['action'] == 'delete':
            for item in patch['data']:
                del self.data[item['id']]
            self.update_snapshot()

        elif patch['action'] == 'insert':
            self.data.update((item['id'], item) for item in patch['data'])
            self.update_snapshot()

    def update_snapshot(self):
        def sortkey(item):
            if item['side'] == 'Sell':
                side = 0
                price = item['price']
            else:
                side = 1
                price = -item['price']
            return side, price

        data = sorted(self.data.values(), key=sortkey)
        self.snapshot['data'] = data



import json
import bisect
import asyncio

import pendulum
from yarl import URL

from uxapi import register_exchange
from uxapi import UXSymbol
from uxapi import UXPatch
from uxapi import Session
from uxapi import WSHandler
from uxapi import Awaitables
from uxapi.exchanges.ccxt.binance import binance
from uxapi.helpers import deep_extend, is_sorted, contract_delivery_time


@register_exchange('binance')
class Binance(UXPatch, binance):
    def __init__(self, market_type, config=None):
        return super().__init__(market_type, deep_extend({
            'options': {
                'defaultType': market_type,
            }
        }, config or {}))

    def describe(self):
        return self.deep_extend(super().describe(), {
            'deliveryHourUTC': 8,

            'urls': {
                'wsapi': {
                    'market': 'wss://stream.binance.com:9443/stream',
                    'private': 'wss://stream.binance.com:9443/ws',
                    'dapiMarket': 'wss://dstream.binance.com/stream',
                    'dapiPrivate': 'wss://dstream.binance.com/ws',
                    'fapiMarket': 'wss://fstream.binance.com/stream',
                    'fapiPrivate': 'wss://fstream.binance.com/ws',
                }
            },

            'wsapi': {
                'market': {
                    'orderbook': '{symbol}@depth{level_speed}',
                    'ohlcv': '{symbol}@kline_{period}',
                    'trade': '{symbol}@trade',
                    'aggTrade': '{symbol}@aggTrade',
                    'ticker': '{symbol}@ticker',
                    '!ticker': '!ticker@arr',
                    'miniTicker': '{symbol}@miniTicker',
                    '!miniTicker': '!miniTicker@arr',
                    'quote': '{symbol}@bookTicker',
                    '!quote': '!bookTicker',
                },
                'private': {'private': 'private'},
                'dapiMarket': {
                    'aggTrade': '{symbol}@aggTrade',
                    'indexPrice': '{pair}@indexPrice{speed_in_seconds}',
                    'markPrice': '{symbol}@markPrice{speed_in_seconds}',
                    'markPriceOfPair': '{pair}@markPrice{speed_in_seconds}',
                    'ohlcv': '{symbol}@kline_{period}',
                    'continuousKline': '{pair_contractType}@continuousKline_{period}',
                    'indexPriceKline': '{pair}@indexPriceKline_{period}',
                    'markPriceKline': '{symbol}@markPriceKline_{period}',
                    'miniTicker': '{symbol}@miniTicker',
                    '!miniTicker': '!miniTicker@arr',
                    'ticker': '{symbol}@ticker',
                    '!ticker': '!ticker@arr',
                    'quote': '{symbol}@bookTicker',
                    '!quote': '!bookTicker',
                    'forceOrder': '{symbol}@forceOrder',
                    '!forceOrder': '!forceOrder@arr',
                    'orderbook': '{symbol}@depth{level_speed}',
                },
                'dapiPrivate': {'private': 'private'},
                'fapiMarket': {
                    'orderbook': '{symbol}@depth{level_speed}',
                    'ohlcv': '{symbol}@kline_{period}',
                    'continuousKline': '{pair_contractType}@continuousKline_{period}',
                    'aggTrade': '{symbol}@aggTrade',
                    'markPrice': '{symbol}@markPrice{speed_in_seconds}',
                    '!markPrice': '!markPrice@arr@{speed_in_seconds}',
                    'miniTicker': '{symbol}@miniTicker',
                    'ticker': '{symbol}@ticker',
                    'quote': '{symbol}@bookTicker',
                    '!quote': '!bookTicker',
                },
                'fapiPrivate': {'private': 'private'},
            }
        })

    def _fetch_markets(self, params=None):
        markets = super()._fetch_markets(params)
        for market in markets:
            if market['type'].startswith(('futures', 'swap')):
                if market['type'].endswith('.usdt'):
                    market['contractValue'] = 1
                else:
                    market['contractValue'] = market['info'].get('contractSize')

                if market['type'].startswith('futures'):
                    timestamp = self.safe_integer(market['info'], 'deliveryDate')
                    delivery_time = pendulum.from_timestamp(timestamp / 1000)
                    market['deliveryTime'] = delivery_time.to_iso8601_string()
        return markets

    def order_book_merger(self):
        return BinanceOrderBookMerger(self)

    def wshandler(self, topic_set):
        wsapi_types = set(self.wsapi_type(topic) for topic in topic_set)
        if len(wsapi_types) > 1:
            raise ValueError('invalid topic_set')
        wsapi_type = wsapi_types.pop()
        wsurl = self.urls['wsapi'][wsapi_type]
        return BinanceWSHandler(self, wsurl, topic_set, wsapi_type)

    def wsapi_type(self, uxtopic):
        if uxtopic.maintype == 'private':
            wstype = 'private'
        else:
            wstype = 'market'

        market_type = uxtopic.market_type
        if market_type.startswith(('futures', 'swap')):
            if market_type.endswith('.usdt'):
                api = 'fapi'
            else:
                api = 'dapi'
            return f'{api}{wstype[0].upper()}{wstype[1:]}'
        else:
            return wstype

    def convert_symbol(self, uxsymbol):
        if uxsymbol.market_type in ('spot', 'margin'):
            return uxsymbol.name

        market_type = uxsymbol.market_type
        base, quote = uxsymbol.base_quote
        if market_type.endswith('.usdt'):
            base, quote = quote, base

        if market_type.startswith('futures'):
            delivery_time = contract_delivery_time(
                expiration=uxsymbol.contract_expiration,
                delivery_hour=self.deliveryHourUTC
            )
            return f'{base}{quote}_{delivery_time:%y%m%d}'
        
        if market_type == 'swap':
            return f'{base}{quote}_PERP'
        elif market_type == 'swap.usdt':
            return f'{base}{quote}'

        raise ValueError(f'invalid symbol: {uxsymbol}')

    def convert_topic(self, uxtopic):
        maintype = uxtopic.maintype
        subtypes = uxtopic.subtypes
        wsapi_type = self.wsapi_type(uxtopic)
        template = self.wsapi[wsapi_type][maintype]

        params = {}
        if uxtopic.extrainfo:
            exchange_id, market_type, _, extrainfo = uxtopic
            uxsymbol = UXSymbol(exchange_id, market_type, extrainfo)
            if maintype == 'continuousKline':
                params['pair_contractType'] = extrainfo
            elif maintype in ('indexPrice', 'markPriceOfPair', 'indexPriceKline'):
                params['pair'] = extrainfo
            else:
                params['symbol'] = self.market_id(uxsymbol).lower()

        if maintype == 'orderbook':
            if not subtypes:
                level_speed = '20@100ms'
            elif subtypes[0] == 'full':
                level_speed = '@0ms' if self.market_type == 'swap' else '@100ms'
            else:
                level_speed = subtypes[0]
            params['level_speed'] = level_speed

        if maintype in ('ohlcv', 'continuousKline', 'indexPriceKline', 'markPriceKline'):
            if subtypes:
                params['period'] = self.timeframes[subtypes[0]]
            else:
                params['period'] = '1m'

        if maintype in ('indexPrice', 'markPrice', '!markPrice'):
            if subtypes and subtypes[0] == '1s':
                speed_in_seconds = '@1s'
            else:
                speed_in_seconds = ''
            params['speed_in_seconds'] = speed_in_seconds

        return template.format(**params)


class BinanceWSHandler(WSHandler):
    def __init__(self, exchange, wsurl, topic_set, wsapi_type):
        super().__init__(exchange, wsurl, topic_set)
        self.wsapi_type = wsapi_type
        self.listen_key = None

    async def connect(self):
        if not self.session:
            self.session = Session()
            self.own_session = True

        if self.login_required:
            result = await self.request_listen_key('POST')
            self.listen_key = result['listenKey']
            wsurl = URL(f'{self.wsurl}/{self.listen_key}')
        else:
            topic_set = [self.convert_topic(topic) for topic in self.topic_set]
            query = {'streams': '/'.join(topic_set)}
            wsurl = URL(self.wsurl).with_query(query)
        ws = await self.session.ws_connect(str(wsurl))
        return ws

    async def request_listen_key(self, method, params=None):
        if self.wsapi_type == 'private':
            baseurl = self.exchange.urls['api']['private']
            if self.exchange.market_type == 'margin':
                baseurl = self.exchange.urls['api']['sapi']
            url = f'{baseurl}/userDataStream'
        elif self.wsapi_type in ('dapiPrivate', 'fapiPrivate'):
            baseurl = self.exchange.urls['api'][self.wsapi_type]
            url = f'{baseurl}/listenKey'
        else:
            raise RuntimeError(f'invalid wsapi_type: {self.wsapi_type}')
        credentials = self.get_credentials()
        headers = {'X-MBX-APIKEY': credentials['apiKey']}
        async with self.session.request(method, url, params=params, headers=headers) as resp:
            result = await resp.json()
        return result

    @property
    def login_required(self):
        return 'private' in self.wsapi_type.lower()

    def prepare(self):
        if self.login_required:
            self.awaitables.create_task(self.keepalive(), 'keepalive')

    async def keepalive(self):
        interval = 20 * 60
        while True:
            await asyncio.sleep(interval)
            params = None
            if self.wsapi_type == 'private':
                params = {'listenKey': self.listen_key}
            try:
                await self.request_listen_key('PUT', params)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger.exception('request listen key failed')

    def decode(self, data):
        return json.loads(data)


class BinanceOrderBookMerger:
    def __init__(self, exchange):
        self.exchange = exchange
        self.snapshot = None
        self.cache = []
        self.future = None
        self.prices = None

    def __call__(self, patch):
        if self.snapshot:
            self.merge(patch)
            return self.snapshot

        self.cache.append(patch)
        if not self.future:
            self.future = Awaitables.default().run_in_executor(
                self.fetch_order_book, patch['data']['s'])
        if not self.future.done():
            raise StopIteration

        snapshot = self.future.result()
        self.future = None
        array = [patch['data']['u'] for patch in self.cache]
        i = bisect.bisect_right(array, snapshot['lastUpdateId'])
        self.cache = self.cache[i:]
        self.on_snapshot(snapshot)
        return self.snapshot

    def on_snapshot(self, snapshot):
        self.snapshot = snapshot
        self.snapshot['lastUpdateId'] = None
        self.prices = {
            'asks': [float(item[0]) for item in snapshot['asks']],
            'bids': [-float(item[0]) for item in snapshot['bids']]
        }
        assert is_sorted(self.prices['asks'])
        assert is_sorted(self.prices['bids'])
        for patch in self.cache:
            self.merge(patch)
        self.cache = None

    def merge(self, patch):
        data = patch['data']
        lastUpdateId = self.snapshot['lastUpdateId']
        if lastUpdateId:
            if 'pu' in data:   # binance futures
                if data['pu'] != lastUpdateId:
                    raise ValueError('invalid patch')
            else:   # binance spot
                if data['U'] != lastUpdateId + 1:
                    raise ValueError('invalid patch')
        self.snapshot['lastUpdateId'] = data['u']
        self.merge_asks_bids(self.snapshot['asks'], data['a'],
                             self.prices['asks'], False)
        self.merge_asks_bids(self.snapshot['bids'], data['b'],
                             self.prices['bids'], True)

    def merge_asks_bids(self, snapshot_lst, patch_lst, price_lst, negative_price):
        for item in patch_lst:
            price, amount = float(item[0]), float(item[1])
            if negative_price:
                price = -price
            i = bisect.bisect_left(price_lst, price)
            if i != len(price_lst) and price_lst[i] == price:
                if amount == 0:
                    price_lst.pop(i)
                    snapshot_lst.pop(i)
                else:
                    snapshot_lst[i] = item
            else:
                if amount != 0:
                    price_lst.insert(i, price)
                    snapshot_lst.insert(i, item)

    def fetch_order_book(self, market_id):
        params = {
            'symbol': market_id,
            'limit': 1000,
        }
        market = self.exchange.markets_by_id[market_id]
        method = self.exchange.find_method(market['type'], 'publicGetDepth')
        return method(params)



from uxapi.__version__ import VERSION, __version__
from uxapi.symbol import UXSymbol
from uxapi.topic import UXTopic
from uxapi.pipeline import Pipeline
from uxapi.listiter import listiter
from uxapi.event import Event
from uxapi.queue import Queue
from uxapi.session import Session
from uxapi.awaitables import (
    Awaitables,
    run_in_executor,
    ExecutionResult,
    ExecutionError,
)
from uxapi.patch import UXPatch
from uxapi.wshandler import WSHandler


_registry = {}


def register_exchange(exchange_id):
    def register(exchange_class):
        _registry[exchange_id] = exchange_class
        exchange_class.id = exchange_id
        return exchange_class
    return register


def new_exchange(exchange_id, market_type, config=None):
    return _registry[exchange_id](market_type, config)


from uxapi.exchanges.okex import (
    Okex, OkexWSHandler, OkexOrderBookMerger)
from uxapi.exchanges.huobi import (
    Huobi, HuobiWSHandler, HuobiWSReq,
    Huobipro, HuobiproOrderBookMerger,
    Huobidm, HuobidmOrderBookMerger)
from uxapi.exchanges.bitmex import (
    Bitmex, BitmexWSHandler, BitmexOrderBookMerger)
from uxapi.exchanges.binance import (
    Binance, BinanceWSHandler, BinanceOrderBookMerger)
from uxapi.exchanges.deribit import (
    Deribit
)



import asyncio


class Event:
    def __init__(self):
        self._event_obj = None

    @property
    def event_obj(self):
        if self._event_obj is None:
            loop = asyncio.get_running_loop()
            self._event_obj = asyncio.Event(loop=loop)
        return self._event_obj

    def __getattr__(self, attr):
        return getattr(self.event_obj, attr)



import ccxt
import pendulum

from uxapi import register_exchange
from uxapi import UXPatch
from uxapi.helpers import contract_delivery_time


@register_exchange('deribit')
class Deribit(UXPatch, ccxt.deribit):
    def describe(self):
        return self.deep_extend(super().describe(), {
            'deliveryHourUTC': 8,
        })

    def _fetch_markets(self, params):
        markets = super()._fetch_markets(params)
        for market in markets:
            market['contractValue'] = self.safe_float(market['info'], 'contract_size')
            if market['info']['settlement_period'] != 'perpetual':
                timestamp = self.safe_integer(market['info'], 'expiration_timestamp')
                delivery_time = pendulum.from_timestamp(timestamp / 1000)
                market['deliveryTime'] = delivery_time.to_iso8601_string()
        return markets

    def convert_symbol(self, uxsymbol):
        month_names = [
            '', 'JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN',
            'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC',
        ]

        if uxsymbol.market_type == 'futures':
            dt = contract_delivery_time(
                expiration=uxsymbol.contract_expiration,
                delivery_hour=self.deliveryHourUTC)
            return f'{uxsymbol.base}-{dt:%d}{month_names[dt.month]}{dt:%y}'

        elif uxsymbol.market_type == 'swap':
            return f'{uxsymbol.base}-PERPETUAL'

        elif uxsymbol.market_type == 'option':
            dt = pendulum.parse(uxsymbol.contract_expiration)
            return (
                f'{uxsymbol.base}-{dt.day}{month_names[dt.month]}{dt:%y}-'
                f'{uxsymbol.option_strike_price}-{uxsymbol.option_type}'
            )

        raise ValueError('invalid symbol')



import functools
import asyncio
from typing import NamedTuple, Optional

from uxapi import Event


class ExecutionError(Exception):
    def __init__(self, name):
        self.name = name


class ExecutionResult(NamedTuple):
    name: Optional[str]
    result: object


class Awaitables:
    @staticmethod
    def default():
        return _default

    def __init__(self):
        self.aws = {}
        self.aws_changed = Event()

    def __iter__(self):
        return iter(self.aws)

    def __len__(self):
        return len(self.aws)

    def __contains__(self, aw_or_name):
        if isinstance(aw_or_name, str):
            name = aw_or_name
            return bool(name and name in self.aws.values())
        else:
            aw = aw_or_name
            return aw in self.aws

    def get_name(self, aw):
        return self.aws[aw]

    def add(self, aw, name=None):
        if aw in self.aws:
            raise ValueError('awaitable object already exist')
        if name and name in self.aws.values():
            raise ValueError('name already exist')
        self.aws[aw] = name
        self.aws_changed.set()

    def create_task(self, coro, name=None):
        task = asyncio.create_task(coro)
        self.add(task, name)
        return task

    def run_in_executor(self, func, *args, executor=None, name=None):
        loop = asyncio.get_running_loop()
        fut = loop.run_in_executor(executor, func, *args)
        self.add(fut, name)
        return fut

    def _cancel_all(self):
        for task in self.aws:
            task.cancel()

    async def cleanup(self):
        if not self.aws:
            return
        self._cancel_all()
        try:
            await asyncio.wait(
                self.aws.keys(),
                return_when=asyncio.ALL_COMPLETED)
        except Exception:
            pass
        self.aws.clear()

    async def wait(self, timeout=None):
        while True:
            self.aws_changed.clear()
            wait_aws_change = asyncio.create_task(self.aws_changed.wait())
            try:
                done, _ = await asyncio.wait(
                    self.aws.keys() | {wait_aws_change},
                    timeout=timeout,
                    return_when=asyncio.FIRST_COMPLETED)
            except asyncio.CancelledError:
                await self.cleanup()
                raise
            finally:
                wait_aws_change.cancel()
            if wait_aws_change not in done:
                break

        if not done:
            raise asyncio.TimeoutError()

        task = done.pop()
        name = self.aws[task]
        del self.aws[task]
        try:
            res = task.result()
        except Exception as exc:
            raise ExecutionError(name) from exc
        return ExecutionResult(name, res)


_default = Awaitables()


def run_in_executor(func, executor=None, name=None):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return Awaitables.default().run_in_executor(
            functools.partial(func, *args, **kwargs),
            executor=executor,
            name=name)
    return wrapper



import time
import hmac as _hmac
import base64
import collections
import itertools
from operator import itemgetter
import pendulum


def current_timestamp():
    return int(time.time()) * 1000


def is_sorted(iterable, *, key=None, reverse=False):
    it1, it2 = itertools.tee(iterable)
    next(it2, None)
    if reverse:
        it1, it2 = it2, it1
    for a, b in zip(it1, it2):
        if key:
            a, b = key(a), key(b)
        if a > b:
            return False
    return True


def all_equal(lst):
    return lst[1:] == lst[:-1]


def extend(*args):
    result = {}
    if args:
        if isinstance(type(args[0]), collections.OrderedDict):
            result = collections.OrderedDict()
        for arg in args:
            if arg:
                result.update(arg)
    return result


def deep_extend(*args):
    result = None
    for arg in args:
        if isinstance(arg, dict):
            if not isinstance(result, dict):
                result = {}
            for key in arg:
                result[key] = deep_extend(result.get(key), arg[key])
        else:
            result = arg
    return result


def keysort(d):
    return collections.OrderedDict(sorted(d.items(), key=itemgetter(0)))


def to_timestamp(text, **options):
    dt = pendulum.parse(text, **options)
    return dt.timestamp() * 1000


def hmac(secret, msg, algorithm='sha256', digest='base64'):
    h = _hmac.new(secret, msg, algorithm)
    if digest == 'hex':
        return h.hexdigest()
    elif digest == 'base64':
        return base64.b64encode(h.digest())
    return h.digest()


def contract_delivery_time(expiration, delivery_hour, since=None):
    """返回合约交割时间

    :param expiration: 当周合约('CW')，次周合约('NW'), 季度合约('CQ')
    :param delivery_hour: 合约交割时间(UTC)
    :param since: 参照时间(UTC), datetime类型
    :return: 合约交割时间, datetime类型
    """
    since = since or pendulum.now('UTC')
    since = pendulum.instance(since)

    if expiration == 'CW':
        cw = since.start_of('week').add(days=4).add(hours=delivery_hour)
        if since > cw:
            cw = cw.next(pendulum.FRIDAY, keep_time=True)
        return cw

    if expiration == 'NW':
        cw = contract_delivery_time('CW', delivery_hour, since)
        return cw.next(pendulum.FRIDAY, keep_time=True)

    if expiration == 'CQ':
        last_friday = since.last_of('quarter', pendulum.FRIDAY)
        last_friday = last_friday.add(hours=delivery_hour)
        if since >= last_friday.subtract(weeks=2):
            since = start_of('next_quarter', since)
            return contract_delivery_time('CQ', delivery_hour, since)
        else:
            return last_friday

    if expiration == 'NQ':
        cq = contract_delivery_time('CQ', delivery_hour, since)
        since = start_of('next_quarter', cq)
        return contract_delivery_time('CQ', delivery_hour, since)

    raise ValueError('invalid expiration')


_PENDULUM_UNITS = ['day', 'week', 'month', 'year', 'decade', 'century']
_EXTENDED_UNITS = [
    'previous_week', 'next_week',
    'previous_month', 'next_month',
    'previous_quarter', 'quarter', 'next_quarter',
]


def start_of(unit, dt):
    if unit not in (_PENDULUM_UNITS + _EXTENDED_UNITS):
        raise ValueError(f'Invalid unit "{unit}" for start_of()')
    dt = pendulum.instance(dt)
    if unit in _PENDULUM_UNITS:
        return dt.start_of(unit)
    if unit == 'quarter':
        start_month = (dt.quarter - 1) * 3 + 1
        return dt.set(month=start_month).start_of('month')
    previous_or_next, unit = unit.split('_')
    if previous_or_next == 'previous':
        previous_day = start_of(unit, dt).subtract(days=1)
        return start_of(unit, previous_day)
    else:
        next_day = end_of(unit, dt).add(days=1)
        return start_of(unit, next_day)


def end_of(unit, dt):
    if unit not in (_PENDULUM_UNITS + _EXTENDED_UNITS):
        raise ValueError(f'Invalid unit "{unit}" for end_of()')
    dt = pendulum.instance(dt)
    if unit in _PENDULUM_UNITS:
        return dt.end_of(unit)
    if unit == 'quarter':
        end_month = dt.quarter * 3
        return dt.set(month=end_month).end_of('month')
    previous_or_next, unit = unit.split('_')
    if previous_or_next == 'previous':
        previous_day = start_of(unit, dt).subtract(days=1)
        return end_of(unit, previous_day)
    else:
        next_day = end_of(unit, dt).add(days=1)
        return end_of(unit, next_day)



import asyncio


class Queue:
    def __init__(self):
        self._queue_obj = None

    @property
    def queue_obj(self):
        if self._queue_obj is None:
            loop = asyncio.get_running_loop()
            self._queue_obj = asyncio.Queue(loop=loop)
        return self._queue_obj

    def __getattr__(self, attr):
        return getattr(self.queue_obj, attr)



import hashlib
from decimal import Decimal

from ccxt.base.exchange import Exchange
from ccxt.base.errors import (
    ExchangeError,
    AuthenticationError,
    ArgumentsRequired,
    OrderNotFound,
    DDoSProtection,
    BadRequest,
    CancelPending,
    InsufficientFunds,
    PermissionDenied
)


class huobidm(Exchange):
    def describe(self):
        return self.deep_extend(super(huobidm, self).describe(), {
            'id': 'huobidm',
            'name': 'Huobi DM',
            'countries': ['CN'],
            'rateLimit': 50,
            'userAgent': self.userAgents['chrome39'],
            'version': 'v1',
            'hostname': 'api.hbdm.com',
            'has': {
                'cancelAllOrders': True,
                'createMarketOrder': 'emulated',
                'fetchMyTrades': True,
                'fetchOHLCV': True,
                'fetchOrder': True,
                'fetchOrders': True,
                'fetchOpenOrders': True,
                'fetchStatus': True,
            },
            'timeframes': {
                '1m': '1min',
                '5m': '5min',
                '15m': '15min',
                '30m': '30min',
                '1h': '60min',
                '4h': '4hour',
                '1d': '1day',
                '1w': '1week',
                '1M': '1mon',
            },
            'expirations': {
                'this_week': 'CW',
                'next_week': 'NW',
                'quarter': 'CQ',
                'next_quarter': 'NQ',
            },
            'urls': {
                'logo': 'https://user-images.githubusercontent.com/4180036/64329037-1d804100-d001-11e9-987f-43c114fab34e.png',
                'api': 'https://api.hbdm.com',
                'heartbeat': 'https://www.hbdm.com',
                'www': 'https://www.hbdm.com',
                'doc': 'https://huobiapi.github.io/docs/dm/v1/cn/',
                'fees': 'https://support.huobi.so/hc/en-us/articles/360000113122',
            },

            'api': {
                'futures': {
                    'get': [
                        'contract_info',                # 获取合约信息
                        'index',                        # 获取合约指数信息
                        'price_limit',                  # 获取合约最高限价和最低限价
                        'open_interest',                # 获取当前可用合约总持仓量
                        'delivery_price',               # 获取预估交割价
                        'api_state',                    # 查询系统状态
                        'market/depth',                 # 获取行情深度数据
                        'market/history/kline',         # 获取K线数据
                        'market/detail/merged',         # 获取聚合行情
                        'market/trade',                 # 获取市场最近成交记录
                        'market/history/trade',         # 批量获取最近的交易记录
                        'risk_info',                    # 查询合约风险准备金余额和预估分摊比例
                        'insurance_fund',               # 查询合约风险准备金余额历史数据
                        'adjustfactor',                 # 查询平台阶梯调整系数
                        'his_open_interest',            # 平台持仓量的查询
                        'elite_account_ratio',          # 精英账户多空持仓对比-账户数
                        'elite_position_ratio',         # 精英账户多空持仓对比-持仓量
                        'liquidation_orders',           # 获取强平单
                        'index/market/history/index',   # 获取指数K线数据
                        'index/market/history/basis',   # 获取基差数据
                        'api_trading_status',           # 获取用户的API指标禁用信息(private)
                    ],

                    'post': [
                        'account_info',                 # 获取用户账户信息
                        'position_info',                # 获取用户持仓信息
                        'sub_account_list',             # 查询母账户下所有子账户资产信息
                        'sub_account_info',             # 查询单个子账户资产信息
                        'sub_position_info',            # 查询单个子账户持仓信息
                        'financial_record',             # 查询用户财务记录
                        'financial_record_exact',       # 组合查询用户财务记录
                        'user_settlement_records',      # 查询用户结算记录
                        'order_limit',                  # 查询用户当前的下单量限制
                        'fee',                          # 查询用户当前的手续费费率
                        'transfer_limit',               # 查询用户当前的划转限制
                        'position_limit',               # 用户持仓量限制的查询
                        'account_position_info',        # 查询用户账户和持仓信息
                        'master_sub_transfer',          # 母子账户划转
                        'master_sub_transfer_record',   # 获取母账户下的所有母子账户划转记录
                        'available_level_rate',         # 查询用户可用杠杆倍数

                        'order',                        # 合约下单
                        'batchorder',                   # 合约批量下单
                        'cancel',                       # 撤销订单
                        'cancelall',                    # 全部撤单
                        'switch_lever_rate',            # 切换杠杆
                        'order_info',                   # 获取合约订单信息
                        'order_detail',                 # 获取订单明细信息
                        'openorders',                   # 获取合约当前未成交委托
                        'hisorders',                    # 获取合约历史委托
                        'hisorders_exact',              # 组合查询合约历史委托
                        'matchresults',                 # 获取历史成交记录
                        'matchresults_exact',           # 组合查询历史成交记录接口
                        'lightning_close_position',     # 闪电平仓下单
                        'trigger_order',                # 合约计划委托下单
                        'trigger_cancel',               # 合约计划委托撤单
                        'trigger_cancelall',            # 合约计划委托全部撤单
                        'trigger_openorders',           # 获取计划委托当前委托
                        'trigger_hisorders',            # 获取计划委托历史委托
                    ]
                },

                'swap': {
                    'get': [
                        'contract_info',                # 获取合约信息
                        'index',                        # 获取合约指数信息
                        'price_limit',                  # 获取合约最高限价和最低限价
                        'open_interest',                # 获取当前可用合约总持仓量
                        'market/depth',                 # 获取行情深度数据
                        'market/history/kline',         # 获取K线数据
                        'market/detail/merged',         # 获取聚合行情
                        'market/trade',                 # 获取市场最近成交记录
                        'market/history/trade',         # 批量获取最近的交易记录
                        'risk_info',                    # 查询合约风险准备金余额和预估分摊比例
                        'insurance_fund',               # 查询合约风险准备金余额历史数据
                        'adjustfactor',                 # 查询平台阶梯调整系数
                        'his_open_interest',            # 平台持仓量的查询
                        'elite_account_ratio',          # 精英账户多空持仓对比-账户数
                        'elite_position_ratio',         # 精英账户多空持仓对比-持仓量
                        'api_state',                    # 查询系统状态
                        'funding_rate',                 # 获取合约的资金费率
                        'historical_funding_rate',      # 获取合约的历史资金费率
                        'liquidation_orders',           # 获取强平订单
                        'index/market/history/swap_premium_index_kline',    # 获取合约的溢价指数K线
                        'index/market/history/swap_estimated_rate_kline',   # 获取实时预测资金费率的K线数据
                        'index/market/history/swap_basis',                  # 获取基差数据
                        'api_trading_status',           # 获取用户的API指标禁用信息(private)
                    ],

                    'post': [
                        'account_info',                 # 获取用户账户信息
                        'position_info',                # 获取用户持仓信息
                        'account_position_info',        # 查询用户账户和持仓信息
                        'sub_account_list',             # 查询母账户下所有子账户资产信息
                        'sub_account_info',             # 查询单个子账户资产信息
                        'sub_position_info',            # 查询单个子账户持仓信息
                        'financial_record',             # 查询用户财务记录
                        'financial_record_exact',       # 组合查询用户财务记录
                        'user_settlement_records',      # 查询用户结算记录
                        'available_level_rate',         # 查询用户可用杠杆倍数
                        'order_limit',                  # 查询用户当前的下单量限制
                        'fee',                          # 查询用户当前的手续费费率
                        'transfer_limit',               # 查询用户当前的划转限制
                        'position_limit',               # 用户持仓量限制的查询
                        'master_sub_transfer',          # 母子账户划转
                        'master_sub_transfer_record',   # 获取母账户下的所有母子账户划转记录

                        'order',                        # 合约下单
                        'batchorder',                   # 合约批量下单
                        'cancel',                       # 撤销订单
                        'cancelall',                    # 全部撤单
                        'switch_lever_rate',            # 切换杠杆
                        'order_info',                   # 获取合约订单信息
                        'order_detail',                 # 获取订单明细信息
                        'openorders',                   # 获取合约当前未成交委托
                        'hisorders',                    # 获取合约历史委托
                        'hisorders_exact',              # 组合查询合约历史委托
                        'matchresults',                 # 获取历史成交记录
                        'matchresults_exact',           # 组合查询用户历史成交记录
                        'lightning_close_position',     # 闪电平仓下单
                        'trigger_order',                # 合约计划委托下单
                        'trigger_cancel',               # 合约计划委托撤单
                        'trigger_cancelall',            # 合约计划委托全部撤单
                        'trigger_openorders',           # 获取计划委托当前委托
                        'trigger_hisorders',            # 获取计划委托历史委托
                    ],
                },

                'swapusdt': {
                    'get': [
                        'contract_info',                # 获取合约信息
                        'index',                        # 获取合约指数信息
                        'price_limit',                  # 获取合约最高限价和最低限价
                        'open_interest',                # 获取当前可用合约总持仓量
                        'market/depth',                 # 获取行情深度数据
                        'market/history/kline',         # 获取K线数据
                        'market/detail/merged',         # 获取聚合行情
                        'market/trade',                 # 获取市场最近成交记录
                        'market/history/trade',         # 批量获取最近的交易记录
                        'risk_info',                    # 查询合约风险准备金余额和预估分摊比例
                        'insurance_fund',               # 查询合约风险准备金余额历史数据
                        'adjustfactor',                 # 查询平台阶梯调整系数
                        'his_open_interest',            # 平台持仓量的查询
                        'elite_account_ratio',          # 精英账户多空持仓对比-账户数
                        'elite_position_ratio',         # 精英账户多空持仓对比-持仓量
                        'api_state',                    # 查询系统状态
                        'funding_rate',                 # 获取合约的资金费率
                        'historical_funding_rate',      # 获取合约的历史资金费率
                        'liquidation_orders',           # 获取强平订单
                        'index/market/history/linear_swap_premium_index_kline',     # 获取合约的溢价指数K线
                        'index/market/history/linear_swap_estimated_rate_kline',    # 获取实时预测资金费率的K线数据
                        'index/market/history/linear_swap_basis',                   # 获取基差数据
                        'api_trading_status',           # 获取用户的API指标禁用信息(private)
                    ],

                    'post': [
                        'account_info',                 # 获取用户账户信息(逐仓)
                        'cross_account_info',           # 获取用户账户信息(全仓)
                        'position_info',                # 获取用户持仓信息(逐仓)
                        'cross_position_info',          # 获取用户持仓信息(全仓)
                        'account_position_info',        # 查询用户账户和持仓信息(逐仓)
                        'cross_account_position_info',  # 查询用户账户和持仓信息(全仓)
                        'swap_sub_auth',                # 批量设置子账户交易权限
                        'sub_account_list',             # 查询母账户下所有子账户资产信息(逐仓)
                        'cross_sub_account_list',       # 查询母账户下所有子账户资产信息(全仓)
                        'sub_account_info_list',        # 批量获取子账户资产信息(逐仓)
                        'cross_sub_account_info_list',  # 批量获取子账户资产信息(全仓)
                        'sub_account_info',             # 查询单个子账户资产信息(逐仓)
                        'cross_sub_account_info',       # 查询单个子账户资产信息(全仓)
                        'sub_position_info',            # 查询单个子账户持仓信息(逐仓)
                        'cross_sub_position_info',      # 查询单个子账户持仓信息(全仓)
                        'financial_record',             # 查询用户财务记录
                        'financial_record_exact',       # 组合查询用户财务记录
                        'user_settlement_records',        # 查询用户结算记录(逐仓)
                        'cross_user_settlement_records',  # 查询用户结算记录(全仓)
                        'available_level_rate',         # 查询用户可用杠杆倍数(逐仓)
                        'cross_available_level_rate',   # 查询用户可用杠杆倍数(全仓)
                        'order_limit',                  # 查询用户当前的下单量限制
                        'fee',                          # 查询用户当前的手续费费率
                        'transfer_limit',               # 查询用户当前的划转限制(逐仓)
                        'cross_transfer_limit',         # 查询用户当前的划转限制(全仓)
                        'position_limit',               # 用户持仓量限制的查询(逐仓)
                        'cross_position_limit',         # 用户持仓量限制的查询(全仓)
                        'master_sub_transfer',          # 母子账户划转
                        'master_sub_transfer_record',   # 获取母账户下的所有母子账户划转记录
                        'transfer_inner',               # 同账号不同保证金账户的划转

                        'order',                        # 合约下单(逐仓)
                        'cross_order',                  # 合约下单(全仓)
                        'batchorder',                   # 合约批量下单(逐仓)
                        'cross_batchorder',             # 合约批量下单(全仓)
                        'cancel',                       # 撤销订单(逐仓)
                        'cross_cancel',                 # 撤销订单(全仓)
                        'cancelall',                    # 全部撤单(逐仓)
                        'cross_cancelall',              # 全部撤单(全仓)
                        'switch_lever_rate',            # 切换杠杆(逐仓)
                        'cross_switch_lever_rate',      # 切换杠杆(全仓)
                        'order_info',                   # 获取合约订单信息(逐仓)
                        'cross_order_info',             # 获取合约订单信息(全仓)
                        'order_detail',                 # 获取订单明细信息(逐仓)
                        'cross_order_detail',           # 获取订单明细信息(全仓)
                        'openorders',                   # 获取合约当前未成交委托(逐仓)
                        'cross_openorders',             # 获取合约当前未成交委托(全仓)
                        'hisorders',                    # 获取合约历史委托(逐仓)
                        'cross_hisorders',              # 获取合约历史委托(全仓)
                        'hisorders_exact',              # 组合查询合约历史委托(逐仓)
                        'cross_hisorders_exact',        # 组合查询合约历史委托(全仓)
                        'matchresults',                 # 获取历史成交记录(逐仓)
                        'cross_matchresults',           # 获取历史成交记录(全仓)
                        'matchresults_exact',           # 组合查询用户历史成交记录(逐仓)
                        'cross_matchresults_exact',     # 组合查询用户历史成交记录(全仓)
                        'lightning_close_position',         # 闪电平仓下单(逐仓)
                        'cross_lightning_close_position',   # 闪电平仓下单(全仓)

                        'trigger_order',                # 合约计划委托下单(逐仓)
                        'cross_trigger_order',          # 合约计划委托下单(全仓)
                        'trigger_cancel',               # 合约计划委托撤单(逐仓)
                        'cross_trigger_cancel',         # 合约计划委托撤单(全仓)
                        'trigger_cancelall',            # 合约计划委托全部撤单(逐仓)
                        'cross_trigger_cancelall',      # 合约计划委托全部撤单(全仓)
                        'trigger_openorders',           # 获取计划委托当前委托(逐仓)
                        'cross_trigger_openorders',     # 获取计划委托当前委托(全仓)
                        'trigger_hisorders',            # 获取计划委托历史委托(逐仓)
                        'cross_trigger_hisorders',      # 获取计划委托历史委托(全仓)

                        'tpsl_order',                   # 对仓位设置止盈止损订单(逐仓)
                        'cross_tpsl_order',             # 对仓位设置止盈止损订单(全仓)
                        'tpsl_cancel',                  # 止盈止损订单撤单(逐仓)
                        'cross_tpsl_cancel',            # 止盈止损订单撤单(全仓)
                        'tpsl_cancelall',               # 止盈止损订单全部撤单(逐仓)
                        'cross_tpsl_cancelall',         # 止盈止损订单全部撤单(全仓)
                        'tpsl_openorders',              # 查询止盈止损订单当前委托(逐仓)
                        'cross_tpsl_openorders',        # 查询止盈止损订单当前委托(全仓)
                        'tpsl_hisorders',               # 查询止盈止损订单历史委托(逐仓)
                        'cross_tpsl_hisorders',         # 查询止盈止损订单历史委托(全仓)
                        'relation_tpsl_order',          # 查询开仓单关联的止盈止损订单详情(逐仓)
                        'cross_relation_tpsl_order',    # 查询开仓单关联的止盈止损订单详情(全仓)
                    ],
                },

                'general': {
                    'get': [
                        'heartbeat',        # 查询系统是否可用
                    ]
                },
            },

            'fees': {
                'futures': {
                    'taker': 0.0003,
                    'maker': 0.0002,
                },
                'swap': {
                    'taker': 0.0002,
                    'maker': 0.0004,
                },
            },

            'exceptions': {
                'exact': {
                    '403': AuthenticationError,   # 无效身份
                    '1017': OrderNotFound,        # 查询订单不存在
                    '1030': BadRequest,           # 输入错误
                    '1031': BadRequest,           # 非法的报单来源
                    '1032': DDoSProtection,       # 访问次数超出限制
                    '1033': BadRequest,           # 合约周期字段值错误
                    '1034': BadRequest,           # 报单价格类型字段值错误
                    '1035': BadRequest,           # 报单方向字段值错误
                    '1036': BadRequest,           # 报单开平字段值错误
                    '1037': BadRequest,           # 杠杆倍数不符合要求
                    '1038': BadRequest,           # 报单价格不符合最小变动价
                    '1039': BadRequest,           # 报单价格超出限制
                    '1040': BadRequest,           # 报单数量不合法
                    '1041': BadRequest,           # 报单数量超出限制
                    '1042': BadRequest,           # 超出多头持仓限制
                    '1043': BadRequest,           # 超出多头持仓限制
                    '1044': BadRequest,           # 超出平台持仓限制
                    '1045': BadRequest,           # 杠杆倍数与所持有仓位的杠杆不符合
                    '1047': InsufficientFunds,    # 可用保证金不足
                    '1048': BadRequest,           # 持仓量不足
                    '1050': BadRequest,           # 客户报单号重复
                    '1051': OrderNotFound,        # 没有可撤订单
                    '1052': BadRequest,           # 超出批量数目限制
                    '1061': OrderNotFound,        # 订单不存在，无法撤单
                    '1062': CancelPending,        # 撤单中，无法重复撤单
                    '1065': BadRequest,           # 客户报单号不是整数
                    '1066': BadRequest,           # 字段不能为空
                    '1067': BadRequest,           # 字段不合法
                    '1069': BadRequest,           # 报单价格不合法
                    '1071': BadRequest,           # 订单已撤，无法撤单
                    '1100': PermissionDenied,     # 用户没有开仓权限
                    '1101': PermissionDenied,     # 用户没有平仓权限
                    '1102': PermissionDenied,     # 用户没有入金权限
                    '1103': PermissionDenied,     # 用户没有出金权限
                    '1200': AuthenticationError,  # 登录错误
                },
            },
            'options': {
                'leverage': 10,
                'defaultType': 'futures',
                'marginMode': 'cross', # 'cross' or 'isolated'
            }
        })

    def find_method(self, market_type, method_name):
        api = market_type.replace('.', '')
        return getattr(self, f'{api}{method_name}')

    def fetch_markets(self, params={}):
        defaultType = self.safe_string_2(self.options, 'fetchMarkets', 'defaultType', 'futures')
        type = self.safe_string(params, 'type', defaultType)
        method = self.find_method(type, 'GetContractInfo')
        response = method(params)
        markets = self.safe_value(response, 'data')
        if not markets:
            raise ExchangeError(self.id + f' {method} returned empty response')
        return [self.parse_market(market) for market in markets]

    def parse_market(self, market):
        # {
        #     symbol: "BTC",
        #     contract_code: "BTC190906",
        #     contract_type: "this_week",
        #     contract_size: 100,
        #     price_tick: 0.01,
        #     delivery_date: "20190906",
        #     create_date: "20190823",
        #     contract_status: 1
        # }
        if 'delivery_date' in market:
            market_type = 'futures'
            base = baseId = market['symbol']
            quote = quoteId = 'USD'
            expiration = self.expirations[market['contract_type']]
            symbol = f'{base}_{expiration}'
        else:
            symbol = market['contract_code']
            base, quote = symbol.split('-')
            if quote == 'USDT':
                market_type = 'swap.usdt'
                base, quote = quote, base
            else:
                market_type = 'swap'
            baseId = base
            quoteId = quote

        active = (self.safe_integer(market, 'contract_status') == 1)
        price_tick = Decimal(market['price_tick'])
        exp = price_tick.adjusted()
        mantissa = price_tick.scaleb(-exp)
        if round(mantissa) == 10:
            price_precision = -exp - 1
        else:
            price_precision = -exp
        return {
            'id': symbol,
            'symbol': symbol,
            'base': base,
            'quote': quote,
            'baseId': baseId,
            'quoteId': quoteId,
            'type': market_type,
            'active': active,
            'precision': {
                'amount': 0,
                'price': price_precision,
            },
            'limits': {
                'amount': {
                    'min': 1,
                    'max': None,
                },
                'price': {
                    'min': None,
                    'max': None
                },
                'cost': {
                    'min': None,
                    'max': None
                }
            },
            'info': market,
        }

    def fetch_ticker(self, symbol, params=None):
        self.load_markets()
        market = self.market(symbol)
        method = self.find_method(market['type'], 'GetMarketDetailMerged')
        if market['type'] == 'futures':
            request = {'symbol': market['id']}
        else:
            request = {'contract_code': market['id']}
        response = method(self.extend(request, params or {}))
        return self.parse_ticker(response['tick'], market)

    def parse_ticker(self, ticker, market):
        # {
        #   "amount": "260479.503",
        #   "ask": [10253.4, 829],
        #   "bid": [10253.39, 885],
        #   "close": "10253.4",
        #   "count": 602655,
        #   "high": "10564.07",
        #   "id": 1568019891,
        #   "low": "10154.79",
        #   "open": "10522.66",
        #   "ts": 1568019891626,
        #   "vol": "27176782"
        # }
        timestamp = self.safe_integer(ticker, 'ts')
        bid = None
        ask = None
        bidVolume = None
        askVolume = None
        if 'bid' in ticker:
            if isinstance(ticker['bid'], list):
                bid = self.safe_float(ticker['bid'], 0)
                bidVolume = self.safe_float(ticker['bid'], 1)
        if 'ask' in ticker:
            if isinstance(ticker['ask'], list):
                ask = self.safe_float(ticker['ask'], 0)
                askVolume = self.safe_float(ticker['ask'], 1)
        open_ = self.safe_float(ticker, 'open')
        close = self.safe_float(ticker, 'close')
        change = None
        percentage = None
        average = None
        if (open_ is not None) and (close is not None):
            change = close - open_
            average = self.sum(open_, close) / 2
            percentage = (change / open_) * 100
        baseVolume = self.safe_float(ticker, 'amount')
        quoteVolume = self.safe_float(ticker, 'vol')
        vwap = None
        if baseVolume is not None and quoteVolume is not None and baseVolume > 0:
            vwap = quoteVolume / baseVolume
        return {
            'symbol': market['symbol'],
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'high': self.safe_float(ticker, 'high'),
            'low': self.safe_float(ticker, 'low'),
            'ask': ask,
            'askVolume': askVolume,
            'bid': bid,
            'bidVolume': bidVolume,
            'vwap': vwap,
            'open': open_,
            'close': close,
            'previousClose': None,
            'change': change,
            'percentage': percentage,
            'average': average,
            'baseVolume': baseVolume,
            'quoteVolume': quoteVolume,
            'info': ticker,
        }

    def fetch_order_book(self, symbol, limit=None, params=None):
        self.load_markets()
        market = self.market(symbol)
        method = self.find_method(market['type'], 'GetMarketDepth')
        if market['type'] == 'futures':
            request = {'symbol': market['id']}
        else:
            request = {'contract_code': market['id']}
        request['type'] = 'step0'
        response = method(self.extend(request, params or {}))
        # {
        #   "asks": [
        #     [10266.11, 6],
        #     [10267.12, 8],
        #     [10267.4, 3],
        #     ...
        #   ],
        #   "bids": [
        #     [10266.1, 2185],
        #     [10265, 13],
        #     [10264.16, 169],
        #     ...
        #   ],
        #   "ch": "market.BTC_CW.depth.step0",
        #   "id": 1568021905,
        #   "mrid": 17747560337,
        #   "ts": 1568021905901,
        #   "version": 1568021905
        # }
        if 'tick' in response:
            if not response['tick']:
                raise ExchangeError(self.id + ' fetchOrderBook() returned empty response: ' + self.json(response))
            orderbook = self.safe_value(response, 'tick')
            result = self.parse_order_book(orderbook, orderbook['ts'])
            result['nonce'] = orderbook['version']
            return result
        raise ExchangeError(self.id + ' fetchOrderBook() returned unrecognized response: ' + self.json(response))

    def fetch_ohlcv(self, symbol, timeframe='1m', since=None,
                    limit=1000, params=None):
        self.load_markets()
        market = self.market(symbol)
        method = self.find_method(market['type'], 'GetMarketHistoryKline')
        if market['type'] == 'futures':
            request = {'symbol': market['id']}
        else:
            request = {'contract_code': market['id']}
        request.update({
            'period': self.timeframes[timeframe],
            'size': limit or 1000
        })
        response = method(self.extend(request, params or {}))
        # {
        #   "ch": "market.BTC_CQ.kline.1min",
        #   "data": [
        #     {
        #       "amount": 223.79326383537403,
        #       "close": 10350.82,
        #       "count": 660,
        #       "high": 10351.34,
        #       "id": 1568022540,
        #       "low": 10344.14,
        #       "open": 10344.15,
        #       "vol": 23158
        #     }
        #   ],
        #   "status": "ok",
        #   "ts": 1568022590047
        # }
        return self.parse_ohlcvs(response['data'], market, timeframe, since, limit)

    def parse_ohlcv(self, ohlcv, market=None, timeframe='1m', since=None, limit=None):
        return [
            self.safe_timestamp(ohlcv, 'id'),
            self.safe_float(ohlcv, 'open'),
            self.safe_float(ohlcv, 'high'),
            self.safe_float(ohlcv, 'low'),
            self.safe_float(ohlcv, 'close'),
            self.safe_float(ohlcv, 'amount'),
        ]

    def fetch_trades(self, symbol, since=None, limit=1000, params=None):
        self.load_markets()
        market = self.market(symbol)
        method = self.find_method(market['type'], 'GetMarketHistoryTrade')
        if market['type'] == 'futures':
            request = {'symbol': market['id']}
        else:
            request = {'contract_code': market['id']}
        request['size'] = limit or 1000
        response = method(self.extend(request, params or {}))
        # {
        #   "ch": "market.BTC_CQ.trade.detail",
        #   "data": [
        #     {
        #       "data": [
        #         {
        #           "amount": 20,
        #           "direction": "sell",
        #           "id": 177510101820000,
        #           "price": 10321.21,
        #           "ts": 1568024107339
        #         }
        #       ],
        #       "id": 17751010182,
        #       "ts": 1568024107339
        #     }
        #   ],
        #   "status": "ok",
        #   "ts": 1568024107692
        # }
        data = self.safe_value(response, 'data')
        result = []
        for i in range(0, len(data)):
            trades = self.safe_value(data[i], 'data', [])
            for j in range(0, len(trades)):
                trade = self.parse_trade(trades[j], market)
                result.append(trade)
        result = self.sort_by(result, 'timestamp')
        return self.filter_by_symbol_since_limit(result, symbol, since, limit)

    def fetch_my_trades(self, symbol=None, since=None, limit=None, params=None):
        if symbol is None:
            raise ArgumentsRequired(self.id + ' fetchMyTrades requires a symbol argument')
        self.load_markets()
        market = self.market(symbol)

        if market['type'] == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(market['type'], f'Post{margin_mode}Matchresults')

        if market['type'] == 'futures':
            request = {
                'symbol': market['base'],
                'contract_code': market['info']['contract_code']
            }
        else:
            request = {
                'contract_code': market['id'],
            }
        request.update({
            'trade_type': 0,
            'create_date': 90,
            'page_index': 1,
            'page_size': 50
        })
        response = method(self.extend(request, params or {}))
        # {
        #   "data": {
        #     "current_page": 1,
        #     "total_page": 1,
        #     "total_size": 2,
        #     "trades": [{
        #       "contract_code": "EOS190419",
        #       "contract_type": "this_week",
        #       "create_date": 1555553626736,
        #       "direction": "sell",
        #       "match_id": 3635853382,
        #       "offset": "close",
        #       "offset_profitloss": 0.15646398812252696,
        #       "order_id": 1118,
        #       "symbol": "EOS",
        #       "trade_fee": -0.002897500905469032,
        #       "trade_price": 5.522,
        #       "trade_turnover": 80,
        #       "role": "maker",
        #       "trade_volume": 8
        #     }]
        #   },
        #   "status": "ok",
        #   "ts": 1555654870867
        # }
        return self.parse_trades(response['data']['trades'], market,
                                 since, limit)

    def parse_trade(self, trade, market=None):
        trade_id = self.safe_string_2(trade, 'id', 'match_id')
        order_id = self.safe_string(trade, 'order_id')
        timestamp = self.safe_integer_2(trade, 'ts', 'create_date')
        price = self.safe_float_2(trade, 'price', 'trade_price')
        amount = self.safe_float_2(trade, 'amount', 'trade_volume')
        side = self.safe_string(trade, 'direction')
        role = self.safe_string(trade, 'role')
        fee = None
        fee_cost = self.safe_float(trade, 'trade_fee')
        if fee_cost is not None:
            fee = {
                'cost': -fee_cost,
                'currency': self.safe_string(trade, 'fee_asset'),
            }
        symbol = market['symbol'] if market else None
        return {
            'id': trade_id,
            'info': trade,
            'order': order_id,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': symbol,
            'type': None,
            'side': side,
            'takerOrMaker': role,
            'price': price,
            'amount': amount,
            'cost': None,
            'fee': fee,
        }

    def fetch_order(self, id, symbol=None, params=None):
        params = params or {}
        ids = [id] if id else None
        orders = self.fetch_orders_by_ids(ids, symbol, params=params)
        if orders:
            return orders[0]
        raise OrderNotFound(f'{self.id} order {id} not found')

    def fetch_orders(self, symbol=None, since=None, limit=None, params=None):
        return self.fetch_orders_by_ids(None, symbol, since, limit, params or {})

    def fetch_orders_by_ids(self, ids, symbol, since=None,
                            limit=None, params=None):
        if symbol is None:
            raise ArgumentsRequired(self.id + ' fetchOpenOrders requires a symbol argument')
        self.load_markets()
        market = self.market(symbol)

        if market['type'] == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(market['type'], f'Post{margin_mode}OrderInfo')

        if market['type'] == 'futures':
            request = {'symbol': market['base']}
        else:
            request = {'contract_code': market['id']}
        if ids:
            request['order_id'] = ','.join(ids)
        response = method(self.extend(request, params or {}))
        # {
        #   "status": "ok",
        #   "data": [
        #     {
        #       "symbol": "BTC",
        #       "contract_code": "BTC190927",
        #       "contract_type": "quarter",
        #       "volume": 1,
        #       "price": 10400,
        #       "order_price_type": "limit",
        #       "order_type": 1,
        #       "direction": "buy",
        #       "offset": "open",
        #       "lever_rate": 10,
        #       "order_id": 6145283623,
        #       "client_order_id": null,
        #       "created_at": 1568106770000,
        #       "trade_volume": 1,
        #       "trade_turnover": 100,
        #       "fee": -0.000002414311264596,
        #       "trade_avg_price": 10354.92,
        #       "margin_frozen": 0,
        #       "profit": 0,
        #       "status": 6,
        #       "order_source": "api"
        #     },
        #     ...
        #   ],
        #   "ts": 1568106883922
        # }
        return self.parse_orders(response['data'], market, since, limit)

    def fetch_open_orders(self, symbol=None, since=None,
                          limit=None, params=None):
        if symbol is None:
            raise ArgumentsRequired(self.id + ' fetchOpenOrders requires a symbol argument')
        self.load_markets()
        market = self.market(symbol)

        if market['type'] == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(market['type'], f'Post{margin_mode}Openorders')

        if market['type'] == 'futures':
            request = {'symbol': market['base']}
        else:
            request = {'contract_code': market['id']}
        request.update({
            'page_index': 1,
            'page_size': 50,
        })
        response = method(self.extend(request, params or {}))
        # {
        #   "status": "ok",
        #   "data":{
        #     "orders":[
        #       {
        #          "symbol": "BTC",
        #          "contract_type": "this_week",
        #          "contract_code": "BTC180914",
        #          "volume": 111,
        #          "price": 1111,
        #          "order_price_type": "limit",
        #          "direction": "buy",
        #          "offset": "open",
        #          "lever_rate": 10,
        #          "order_id": 106837,
        #          "client_order_id": 10683,
        #          "order_source": "web",
        #          "created_at": 1408076414000,
        #          "trade_volume": 1,
        #          "trade_turnover": 1200,
        #          "fee": 0,
        #          "trade_avg_price": 10,
        #          "margin_frozen": 10,
        #          "status": 1
        #       }
        #     ],
        #     "total_page":15,
        #     "current_page":3,
        #     "total_size":3
        #   },
        #   "ts": 1490759594752
        # }
        orders = self.parse_orders(response['data']['orders'], market, since, None)
        open_orders = [order for order in orders if order.get('status') == 'open']
        if limit:
            open_orders = open_orders[0:limit]
        return open_orders

    def parse_order(self, order, market=None):
        timestamp = self.safe_integer(order, 'created_at')
        status = self.safe_string(order, 'status')
        status = self.parse_order_status(status)
        currency = self.safe_string(order, 'symbol')
        type = market['type'] if market else self.options['defaultType']
        if type == 'futures':
            contract_type = self.safe_string(order, 'contract_type')
            expiration = self.expirations[contract_type]
            symbol = f'{currency}_{expiration}'
        else:
            symbol = self.safe_string(order, 'contract_code')
        side = self.safe_string(order, 'direction')
        price = self.safe_float(order, 'price')
        average = self.safe_float(order, 'trade_avg_price')
        amount = self.safe_float(order, 'volume')
        filled = self.safe_float(order, 'trade_volume')
        remaining = None
        if filled is not None and amount is not None:
            remaining = amount - filled
        fee = None
        fee_cost = self.safe_float(order, 'fee')
        if fee_cost is not None:
            fee = {
                'cost': fee_cost,
                'currency': currency,
            }
        return {
            'id': self.safe_string(order, 'order_id'),
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'status': status,
            'symbol': symbol,
            'type': self.safe_string(order, 'order_price_type'),
            'side': side,
            'price': price,
            'average': average,
            'amount': amount,
            'filled': filled,
            'remaining': remaining,
            'cost': None,
            'fee': fee,
            'info': order,
        }

    def parse_order_status(self, status):
        statuses = {
            '1':  'open',       # pre commit
            '2':  'open',       # pre commit
            '3':  'open',       # committed
            '4':  'open',       # partial filled
            '5':  'canceled',   # partial canceled
            '6':  'closed',     # completed
            '7':  'canceled',   # canceled
            '11': 'canceled',   # cancelling
        }
        return self.safe_string(statuses, status, status)

    def create_order(self, symbol, type, side, amount, price=None, params=None):
        self.load_markets()
        market = self.market(symbol)
        
        if market['type'] == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(market['type'], f'Post{margin_mode}Order')

        params = params or {}
        lever_rate = params.get('lever_rate', self.options['leverage'])
        if type == 'market':
            type = 'optimal_20'
        if market['type'] == 'futures':
            request = {
                'symbol': market['base'],
                'contract_type': market['info']['contract_type'],
                'contract_code': market['info']['contract_code'],
            }
        else:
            request = {
                'contract_code': market['id']
            }
        request.update({
            'volume': self.amount_to_precision(symbol, amount),
            'direction': side,
            'offset': 'open',
            'lever_rate': lever_rate,
            'order_price_type': type
        })
        if type in ['limit', 'post_only', 'fok', 'ioc']:
            request['price'] = self.price_to_precision(symbol, price)
        response = method(self.extend(request, params))
        # {
        #   "status": "ok",
        #   "data": {
        #     "order_id": 6145283619
        #   },
        #   "ts": 1568105905237
        # }
        order_id = self.safe_string(response.get('data', {}), 'order_id')
        timestamp = self.safe_integer(response, 'ts')

        return {
            'info': response,
            'id': order_id,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'status': None,
            'symbol': symbol,
            'type': type,
            'side': side,
            'price': price,
            'amount': amount,
            'filled': None,
            'remaining': None,
            'cost': None,
            'trades': None,
            'fee': None,
        }

    def cancel_order(self, id, symbol=None, params=None):
        if symbol is None:
            raise ArgumentsRequired(self.id + ' cancelOrder requires a symbol argument')
        ids = [id] if id else None
        return self.cancel_order_by_ids(ids, symbol, params)

    def cancel_order_by_ids(self, ids, symbol=None, params=None):
        if symbol is None:
            raise ArgumentsRequired(self.id + ' cancelOrderByIds requires a symbol argument')
        market = self.market(symbol)

        if market['type'] == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(market['type'], f'Post{margin_mode}Cancel')

        if market['type'] == 'futures':
            request = {'symbol': market['base']}
        else:
            request = {'contract_code': market['id']}
        if ids:
            request['order_id'] = ','.join(ids)
        response = method(self.extend(request, params or {}))
        # {
        #   "status": "ok",
        #   "data": {
        #     "errors": [
        #       {
        #         "order_id": "161251",
        #         "err_code": 200417,
        #         "err_msg": "invalid symbol"
        #       },
        #       {
        #         "order_id": 161253,
        #         "err_code": 200415,
        #         "err_msg": "invalid symbol"
        #       }
        #     ],
        #     "successes": [
        #       161256,
        #       1344567
        #     ]
        #   },
        #   "ts": 1490759594752
        # }
        return response

    def cancel_all_orders(self, symbol=None, params=None):
        if symbol is None:
            raise ArgumentsRequired(self.id + ' cancelAllOrders requires a symbol argument')
        market = self.market(symbol)

        if market['type'] == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(market['type'], f'Post{margin_mode}Cancelall')

        if market['type'] == 'futures':
            request = {
                'symbol': market['base'],
                'contract_type': market['info']['contract_type']
            }
        else:
            request = {
                'contract_code': market['id']
            }
        return method(self.extend(request, params or {}))

    def fetch_balance(self, params=None):
        self.load_markets()
        defaultType = self.safe_string_2(self.options, 'fetchBalance', 'defaultType')
        type = self.safe_string(params, 'type', defaultType)
        if type is None:
            raise ArgumentsRequired(
                f"{self.id} fetchBalance requires a type parameter"
                f"(one of 'futures', 'swap', 'swap.usdt')"
            )
        
        if type == 'swap.usdt' and self.options['marginMode'] == 'cross':
            margin_mode = 'Cross'
        else:
            margin_mode = ''
        method = self.find_method(type, f'Post{margin_mode}AccountInfo')
        response = method(params or {})
        return self.parse_balance(response)

    def parse_balance(self, response):
        # futures/swap-coin/swap-usdt isolated mode
        # {
        #   "status": "ok",
        #   "data": [
        #     {
        #        "symbol": "BTC",
        #        "margin_balance": 1,
        #        "margin_position": 0,
        #        "margin_frozen": 3.33,
        #        ...
        #     },
        #     ...
        #   ],
        #   "ts":158797866555
        # }

        # swap-usdt cross mode
        # {
        #     "status":"ok",
        #     "data":[
        #         {
        #             "margin_mode":"cross",
        #             "margin_account":"USDT",
        #             "margin_asset":"USDT",
        #             "margin_balance":0.000000549410817836,
        #             ...
        #             "contract_detail":[
        #                 {
        #                     "symbol":"BTC",
        #                     "contract_code":"BTC-USDT",
        #                     ...
        #                 },
        #                 ...
        #             ]
        #         }
        #     ],
        #     "ts":1606906200680
        # }

        balance = {}
        for item in response['data']:
            currency = self.safe_string_2(item, 'symbol', 'margin_account')
            total = self.safe_float(item, 'margin_balance')
            margin_position = self.safe_float(item, 'margin_position', 0.0)
            margin_frozen = self.safe_float(item, 'margin_frozen', 0.0)
            used = margin_position + margin_frozen
            free = total - used
            balance[currency] = {
                'total': total,
                'used': used,
                'free': free,
            }

        totalAll = {currency: balance[currency]['total'] for currency in balance}
        usedAll = {currency: balance[currency]['used'] for currency in balance}
        freeAll = {currency: balance[currency]['free'] for currency in balance}
        balance['total'] = totalAll
        balance['used'] = usedAll
        balance['free'] = freeAll
        return balance
        

    def fetch_status(self, params=None):
        response = self.generalGetHeartbeat(params or {})
        # {
        #    "status":"ok",
        #    "data":{
        #       "heartbeat":1,
        #       "estimated_recovery_time":null,
        #       "swap_heartbeat":1,
        #       "swap_estimated_recovery_time":null
        #    },
        #    "ts":1557714418033
        # }
        defaultType = self.safe_string_2(self.options, 'fetchStatus', 'defaultType')
        type = self.safe_string(params, 'type', defaultType)
        if type.startswith('swap'):
            key = 'swap_heartbeat'
        else:
            key = 'heartbeat'
        heartbeat = self.safe_integer(response['data'], key)
        if heartbeat == 1:
            status = 'ok'
        else:
            status = 'maintenance'
        self.status = {
            'status': status,
            'updated': self.milliseconds(),
        }
        return self.status

    def sign(self, path, api='public', method='GET', params={}, headers=None, body=None):
        if api == 'futures':
            if path.startswith(('market/', 'index/', 'lightning_')):
                prefix = ''
            else:
                prefix = f'api/{self.version}/contract_'

        elif api == 'swap':
            if path.startswith('market/'):
                prefix = 'swap-ex/'
            elif path.startswith('index/'):
                prefix = ''
            else:
                prefix = f'swap-api/{self.version}/swap_'

        elif api == 'swapusdt':
            if path.startswith('market/'):
                prefix = 'linear-swap-ex/'
            elif path.startswith('index/'):
                prefix = ''
            else:
                prefix = f'linear-swap-api/{self.version}/swap_'

        else:
            assert api == 'general' and path == 'heartbeat'
            url = self.urls['heartbeat'] + '/heartbeat'
            return {'url': url, 'method': method, 'body': body, 'headers': headers}

        url = f'/{prefix}{path}'
        url = self.implode_params(url, params)
        query = self.omit(params, self.extract_params(path))
        if method == 'POST' or path == 'api_trading_status':
            self.check_required_credentials()
            timestamp = self.ymdhms(self.milliseconds(), 'T')
            request = self.keysort({
                'SignatureMethod': 'HmacSHA256',
                'SignatureVersion': '2',
                'AccessKeyId': self.apiKey,
                'Timestamp': timestamp,
            })
            auth = self.urlencode(request)
            payload = '\n'.join([method, self.hostname, url, auth])
            signature = self.hmac(self.encode(payload), self.encode(self.secret), hashlib.sha256, 'base64')
            auth += '&' + self.urlencode({'Signature': signature})
            url += '?' + auth
            if method == 'POST':
                body = self.json(query)
                headers = {'Content-Type': 'application/json'}
            else:
                headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        else:
            if query:
                url += '?' + self.urlencode(query)
        url = self.urls['api'] + url
        return {'url': url, 'method': method, 'body': body, 'headers': headers}

    def handle_errors(self, httpCode, reason, url, method, headers, body,
                      response, requestHeaders, requestBody):
        if response is None:
            return  # fallback to default error handler
        if 'status' in response:
            # {
            #    "status":"error",
            #    "err-code":"order-limitorder-amount-min-error",
            #    "err-msg":"limit order amount error, min: `0.001`",
            #    "data":null
            # }
            status = self.safe_string(response, 'status')
            if status == 'error':
                code = self.safe_string(response, 'err-code')
                feedback = self.id + ' ' + body
                self.throw_exactly_matched_exception(self.exceptions['exact'], code, feedback)
                message = self.safe_string(response, 'err-msg')
                self.throw_exactly_matched_exception(self.exceptions['exact'], message, feedback)
                raise ExchangeError(feedback)



class UXSymbol:
    def __init__(self, exchange_id, market_type, name):
        self.exchange_id = exchange_id
        self.market_type = market_type
        self.name = name
        self.isspecial = self.name.startswith('!')
        if self.isspecial:
            self.name_info = name[1:].split('.')
        else:
            self.name_info = name.split('.')

    @classmethod
    def fromstring(cls, s):
        return cls(*s.split(':'))

    def __repr__(self):
        args = ', '.join(repr(field) for field in tuple(self))
        return f'UXSymbol({args})'

    def __str__(self):
        return ':'.join(self)

    def __eq__(self, other):
        if isinstance(other, UXSymbol):
            return tuple(self) == tuple(other)
        return False

    def __hash__(self):
        return hash(str(self))

    def __iter__(self):
        return (field for field in (self.exchange_id, self.market_type, self.name))

    @property
    def base_quote(self):
        base, *quote = self.name_info[0].split('/', maxsplit=1)
        if quote:
            return base, quote[0]
        else:
            return base, base

    @property
    def base(self):
        return self.base_quote[0]

    @property
    def quote(self):
        return self.base_quote[1]

    @property
    def contract_expiration(self):
        return self.name_info[1]

    @property
    def option_strike_price(self):
        return self.name_info[2].replace('_', '.')

    @property
    def option_type(self):
        return self.name_info[3]



import asyncio
from yarl import URL
import aiohttp
from aiohttp.client_exceptions import InvalidURL
from aiohttp.helpers import sentinel, proxies_from_env


class Session:
    def __init__(self, **kwargs):
        self._session_obj = None
        self._kwargs = kwargs

    @property
    def session_obj(self):
        if self._session_obj is None:
            if 'loop' not in self._kwargs:
                loop = asyncio.get_running_loop()
                self._kwargs['loop'] = loop
            self._session_obj = _ClientSession(**self._kwargs)
        return self._session_obj

    def __getattr__(self, attr):
        return getattr(self.session_obj, attr)


class _ClientSession(aiohttp.ClientSession):
    def __init__(self, *, trust_env=True, timeout=sentinel, **kwargs):
        if timeout is sentinel:
            timeout = aiohttp.ClientTimeout(total=20)
        return super().__init__(trust_env=trust_env, timeout=timeout, **kwargs)

    async def _request(self, method, str_or_url, *,
                       proxy=None, proxy_auth=None, **kwargs):
        if proxy is None and self._trust_env:
            try:
                url = URL(str_or_url)
            except ValueError:
                raise InvalidURL(str_or_url)
            proxy, proxy_auth = self._proxy_from_env(url.scheme)
        resp = await super()._request(
            method,
            str_or_url,
            proxy=proxy,
            proxy_auth=proxy_auth,
            **kwargs)
        return resp

    @staticmethod
    def _proxy_from_env(scheme):
        if scheme == 'wss':
            scheme = 'https'
        if scheme == 'ws':
            scheme = 'http'
        proxies = proxies_from_env()
        proxy_info = proxies.get(scheme)
        if proxy_info:
            return proxy_info.proxy, proxy_info.proxy_auth
        else:
            return None, None



import datetime
import asyncio
import gzip
import json
import urllib.parse
import bisect
from urllib.parse import parse_qs

import yarl

import pendulum
from ccxt import huobipro
from uxapi import register_exchange
from uxapi import UXSymbol
from uxapi import WSHandler
from uxapi import UXPatch
from uxapi import Queue
from uxapi import Awaitables
from uxapi.exchanges.ccxt.huobidm import huobidm
from uxapi.helpers import (
    keysort,
    hmac,
    extend,
    is_sorted
)


@register_exchange('huobi')
class Huobi:
    def __init__(self, market_type, config):
        if market_type in ('spot', 'margin'):
            cls = Huobipro
        else:
            cls = Huobidm
        cls.id = type(self).id
        self._exchange = cls(market_type, config)

    def __getattr__(self, attr):
        return getattr(self._exchange, attr)


class Huobipro(UXPatch, huobipro):
    def describe(self):
        return self.deep_extend(super().describe(), {
            'has': {
                'fetchCurrencies': True,
                'fetchOrders': True,
                'fetchOpenOrders': True,
                'cancelOrders': True,
                'cancelAllOrders': True,
            },

            'urls': {
                'wsapi': {
                    'market': 'wss://api.huobi.pro/ws',
                    'market_aws': 'wss://api-aws.huobi.pro/ws',
                    'mbp': 'wss://api.huobi.pro/feed',
                    'mbp_aws': 'wss://api-aws.huobi.pro/feed',
                    'private': 'wss://api.huobi.pro/ws/v2',
                    'private_aws': 'wss://api-aws.huobi.pro/ws/v2',
                },
            },

            'wsapi': {
                'market': {
                    'ticker': 'market.{symbol}.detail',
                    'ohlcv': 'market.{symbol}.kline.{period}',
                    'orderbook': 'market.{symbol}.depth.{level}',
                    'refresh': 'market.{symbol}.mbp.refresh.{level}',
                    'trade': 'market.{symbol}.trade.detail',
                    'bbo': 'market.{symbol}.bbo',
                },
                'mbp': {
                    'mbp': 'market.{symbol}.mbp.{level}',
                },
                'private': {
                    'myorder': 'orders#{symbol}',
                    'accounts': 'accounts.update#{mode}',
                    'clearing': 'trade.clearing#{symbol}'
                }
            },
        })

    def fetch_accounts(self, params=None):
        accounts = super().fetch_accounts(params or {})
        if self.market_type != 'margin':
            return accounts
        results = []
        for account in accounts:
            if account['type'] == 'super-margin':
                results.insert(0, account)
            else:
                results.append(account)
        return results

    def order_book_merger(self):
        return HuobiproOrderBookMerger(self)

    def _fetch_markets(self, params=None):
        markets = super()._fetch_markets(params)
        for market in markets:
            market['type'] = 'spot'
        return markets

    def _create_order(self, uxsymbol, type, side, amount, price=None, params=None):
        self.load_accounts()
        account_type = self.accounts[0]['type']
        source = f'{account_type}-api'
        params = self.extend({'source': source}, params or {})
        return super()._create_order(uxsymbol, type, side, amount, price, params)

    def _cancel_orders(self, ids, uxsymbol, params):
        params = params or {}
        request = {
            'order-ids': ids,
        }
        return self.privatePostOrderOrdersBatchcancel(self.extend(request, params))

    def _cancel_all_orders(self, uxsymbol, params):
        params = params or {}
        request = {}
        if not self.safe_string(params, 'account-id'):
            self.loadAccounts()
            request['account-id'] = self.accounts[0]['id']
        if not self.safe_string(params, 'symbol'):
            if uxsymbol:
                request['symbol'] = self.convert_symbol(uxsymbol)
        return self.privatePostOrderOrdersBatchCancelOpenOrders(
            self.extend(request, params))

    def convert_topic(self, uxtopic):
        maintype = uxtopic.maintype
        subtypes = uxtopic.subtypes
        wsapi_type = self.wsapi_type(uxtopic)
        template = self.wsapi[wsapi_type].get(maintype)

        if maintype == 'accounts':
            mode = subtypes[0] if subtypes else '1'
            return template.format(mode=mode)

        params = {}
        uxsymbol = UXSymbol(uxtopic.exchange_id, uxtopic.market_type,
                            uxtopic.extrainfo)
        if uxsymbol.name == '*':
            params['symbol'] = '*'
        else:
            params['symbol'] = self.market_id(uxsymbol)
        if maintype in ['orderbook', 'refresh', 'mbp']:
            if not subtypes:
                assert maintype == 'orderbook'
                params['level'] = 'step0'
            elif subtypes[0] == 'full':
                assert maintype == 'orderbook'
                template = self.wsapi['mbp']['mbp']
                params['level'] = '150'
            else:
                params['level'] = subtypes[0]
        if maintype == 'ohlcv':
            params['period'] = self.timeframes[subtypes[0]]
        return template.format(**params)

    def wshandler(self, topic_set):
        wsapi_types = {self.wsapi_type(topic) for topic in topic_set}
        if len(wsapi_types) > 1:
            raise ValueError('invalid topics')
        wsapi_type = wsapi_types.pop()
        wsurl = self.urls['wsapi'][wsapi_type]
        return HuobiWSHandler(self, wsurl, topic_set, wsapi_type)

    def wsapi_type(self, uxtopic):
        if uxtopic.datatype == 'orderbook.full':
            return 'mbp'

        for type in self.wsapi:
            if uxtopic.maintype in self.wsapi[type]:
                return type

        raise ValueError('invalid topic')


class _HuobiOrderBookMerger:
    def merge_asks_bids(self, snapshot_lst, patch_lst, price_lst, negative_price):
        for item in patch_lst:
            price, amount = item
            if negative_price:
                price = -price
            i = bisect.bisect_left(price_lst, price)
            if i != len(price_lst) and price_lst[i] == price:
                if amount == 0:
                    price_lst.pop(i)
                    snapshot_lst.pop(i)
                else:
                    snapshot_lst[i] = item
            else:
                if amount != 0:
                    price_lst.insert(i, price)
                    snapshot_lst.insert(i, item)


class HuobiproOrderBookMerger(_HuobiOrderBookMerger):
    def __init__(self, exchange):
        self.exchange = exchange
        self.snapshot = None
        self.topic = None
        self.wsreq = None
        self.wsreq_task = None
        self.future = None
        self.cache = []
        self.prices = None

    def __call__(self, patch):
        if self.snapshot:
            self.merge(patch)
            return self.snapshot

        if self.wsreq is None:
            self.topic = patch['ch']
            self.start_wsreq()

        self.cache.append(patch)
        if not self.future:
            self.future = self.wsreq.request({
                'req': self.topic
            })
        if not self.future.done():
            raise StopIteration

        snapshot = self.future.result()
        self.future = None
        seqnums = [item['tick']['prevSeqNum'] for item in self.cache]
        snapshot_seq = snapshot['data']['seqNum']
        assert is_sorted(seqnums)
        i = bisect.bisect_left(seqnums, snapshot_seq)
        if i != len(seqnums) and seqnums[i] == snapshot_seq:
            self.stop_wsreq()
            self.cache = self.cache[i:]
            self.on_snapshot(snapshot)
            return self.snapshot
        raise StopIteration

    def on_snapshot(self, snapshot):
        self.snapshot = {
            'ch': snapshot['rep'],
            'tick': snapshot['data'],
        }
        self.prices = {
            'asks': [item[0] for item in snapshot['data']['asks']],
            'bids': [-item[0] for item in snapshot['data']['bids']]
        }
        for patch in self.cache:
            self.merge(patch)
        self.cache = None

    def merge(self, patch):
        snapshot_tick = self.snapshot['tick']
        patch_tick = patch['tick']
        if snapshot_tick['seqNum'] != patch_tick['prevSeqNum']:
            raise RuntimeError('seqNum error')
        snapshot_tick['seqNum'] = patch_tick['seqNum']
        snapshot_tick['ts'] = patch['ts']
        if 'asks' in patch_tick:
            self.merge_asks_bids(snapshot_tick['asks'], patch_tick['asks'],
                                self.prices['asks'], False)
        if 'bids' in patch_tick:
            self.merge_asks_bids(snapshot_tick['bids'], patch_tick['bids'],
                                self.prices['bids'], True)

    def start_wsreq(self):
        self.wsreq = HuobiWSReq(self.exchange, 'mbp')

        async def run():
            try:
                await self.wsreq.run()
            except asyncio.CancelledError:
                pass

        self.wsreq_task = Awaitables.default().create_task(run())

    def stop_wsreq(self):
        self.wsreq_task.cancel()
        self.wsreq_task = None
        self.wsreq = None


class Huobidm(UXPatch, huobidm):
    def __init__(self, market_type, config=None):
        return super().__init__(market_type, self.deep_extend({
            'options': {'defaultType': market_type}
        }, config or {}))

    def describe(self):
        return self.deep_extend(super().describe(), {
            'deliveryHourUTC': 8,

            'has': {
                'fetchOrders': True,
                'fetchOpenOrders': True,
                'cancelAllOrders': True,
            },

            'urls': {
                'wsapi': {
                    'market': {
                        'futures': 'wss://api.hbdm.com/ws',
                        'swap': 'wss://api.hbdm.com/swap-ws',
                        'swap.usdt': 'wss://api.hbdm.com/linear-swap-ws',
                    },
                    'private': {
                        'futures': 'wss://api.hbdm.com/notification',
                        'swap': 'wss://api.hbdm.com/swap-notification',
                        'swap.usdt': 'wss://api.hbdm.com/linear-swap-notification',
                    },
                    'index': 'wss://api.hbdm.com/ws_index',
                },
            },

            'wsapi': {
                'market': {
                    'ohlcv': 'market.{symbol}.kline.{period}',
                    'orderbook': 'market.{symbol}.depth.{level}',
                    'high_freq': 'market.{symbol}.depth.size_{level}.high_freq?data_type={data_type}',  # noqa: E501
                    'ticker': 'market.{symbol}.detail',
                    'bbo': 'market.{symbol}.bbo',
                    'trade': 'market.{symbol}.trade.detail',
                },
                'private': {
                    'myorder': 'orders.{symbol}',
                    'myorder_cross': 'orders_cross.{symbol}',
                    'accounts': 'accounts.{symbol}',
                    'accounts_cross': 'accounts_cross.{symbol}',
                    'position': 'positions.{symbol}',
                    'position_cross': 'positions_cross.{symbol}',
                    'matchOrders': 'matchOrders.{symbol}',
                    'matchOrders_cross': 'matchOrders_cross.{symbol}',
                    'trigger_order': 'trigger_order.{symbol}',
                    'trigger_order_cross': 'trigger_order_cross.{symbol}',
                },
                'index': {
                    'index': 'market.{symbol}.index.{period}',                     # futures
                    'premium_index': 'market.{symbol}.premium_index.{period}',     # swap
                    'estimated_rate': 'market.{symbol}.estimated_rate.{period}',   # swap
                    'basis': 'market.{symbol}.basis.{period}.{basis_price_type}',
                },
                'public': {
                    'liquidation_orders': 'public.{symbol}.liquidation_orders',
                    'funding_rate': 'public.{symbol}.funding_rate',  # swap
                    'contract_info': 'public.{symbol}.contract_info',
                }
            },
        })

    def order_book_merger(self):
        return HuobidmOrderBookMerger()

    def _fetch_markets(self, params=None):
        markets = super()._fetch_markets(params)
        for market in markets:
            contract_value = self.safe_float(market['info'], 'contract_size')
            market['contractValue'] = contract_value
            if market['type'] == 'futures':
                delivery_date = self.safe_string(market['info'], 'delivery_date')
                if delivery_date:
                    delivery_time = pendulum.from_format(delivery_date, 'YYYYMMDD')
                    delivery_time = delivery_time.add(hours=self.deliveryHourUTC)
                    market['deliveryTime'] = delivery_time.to_iso8601_string()
                else:
                    market['deliveryTime'] = None
        return markets

    def convert_symbol(self, uxsymbol):
        market_type = uxsymbol.market_type
        base, quote = uxsymbol.base_quote
        if market_type == 'futures':
            return f'{base}_{uxsymbol.contract_expiration}'
        elif market_type == 'swap':
            return f'{base}-{quote}'
        elif market_type == 'swap.usdt':
            return f'{quote}-{base}'
        else:
            raise ValueError(f'invalid symbol: {uxsymbol}')

    def convert_topic(self, uxtopic):
        wsapi_type = self.wsapi_type(uxtopic)
        maintype = uxtopic.maintype
        subtypes = uxtopic.subtypes

        if (uxtopic.market_type == 'swap.usdt'
                and self.options['marginMode'] == 'cross'
                and maintype in ('myorder', 'accounts', 'position')):
            maintype += '_cross'

        params = {}
        if wsapi_type in ('private', 'public'):
            params['symbol'] = uxtopic.extrainfo

        elif wsapi_type == 'index':
            params['symbol'] = uxtopic.extrainfo
            params['period'] = self.timeframes[subtypes[0]]
            if maintype == 'basis':
                params['basis_price_type'] = subtypes[1]

        elif wsapi_type == 'market':
            uxsymbol = UXSymbol(uxtopic.exchange_id, uxtopic.market_type,
                                uxtopic.extrainfo)
            params['symbol'] = self.market_id(uxsymbol)
            if maintype == 'orderbook':
                if not subtypes:
                    params['level'] = 'step0'
                elif subtypes[0] == 'full':
                    maintype = 'high_freq'
                    params['level'] = '150'
                    params['data_type'] = 'incremental'
                else:
                    params['level'] = subtypes[0]
            elif maintype == 'high_freq':
                assert subtypes and len(subtypes) == 2
                params['level'] = subtypes[0]
                params['data_type'] = subtypes[1]
            elif maintype == 'ohlcv':
                params['period'] = self.timeframes[subtypes[0]]

        template = self.wsapi[wsapi_type][maintype]
        return template.format(**params)

    def wshandler(self, topic_set):
        wsapi_types = {self.wsapi_type(topic) for topic in topic_set}
        if len(wsapi_types) > 1:
            raise ValueError('invalid topics')
        wsapi_type = wsapi_types.pop()
        if wsapi_type == 'index':
            wsurl = self.urls['wsapi'][wsapi_type]
        elif wsapi_type == 'public':
            wsurl = self.urls['wsapi']['private'][self.market_type]
        else:
            wsurl = self.urls['wsapi'][wsapi_type][self.market_type]
        return HuobiWSHandler(self, wsurl, topic_set, wsapi_type)

    def wsapi_type(self, uxtopic):
        for key in self.wsapi:
            if uxtopic.maintype in self.wsapi[key]:
                return key
        raise ValueError('invalid topic')


class HuobidmOrderBookMerger(_HuobiOrderBookMerger):
    def __init__(self):
        self.snapshot = None
        self.prices = None

    def __call__(self, patch):
        if patch['tick']['event'] == 'snapshot':
            self.on_snapshot(patch)
        elif patch['tick']['event'] == 'update':
            self.merge(patch)
        else:
            raise ValueError('unexpected event')
        return self.snapshot

    def on_snapshot(self, snapshot):
        self.snapshot = snapshot
        self.prices = {
            'asks': [item[0] for item in snapshot['tick']['asks']],
            'bids': [-item[0] for item in snapshot['tick']['bids']]
        }

    def merge(self, patch):
        self.snapshot['ts'] = patch['ts']
        snapshot_tick = self.snapshot['tick']
        patch_tick = patch['tick']
        if snapshot_tick['version'] + 1 != patch_tick['version']:
            raise RuntimeError('version error')
        snapshot_tick.update({
            'mrid': patch_tick['mrid'],
            'id': patch_tick['id'],
            'ts': patch_tick['ts'],
            'version': patch_tick['version'],
        })
        self.merge_asks_bids(snapshot_tick['asks'], patch_tick['asks'],
                             self.prices['asks'], False)
        self.merge_asks_bids(snapshot_tick['bids'], patch_tick['bids'],
                             self.prices['bids'], True)


class HuobiWSHandler(WSHandler):
    def __init__(self, exchange, wsurl, topic_set, wsapi_type):
        super().__init__(exchange, wsurl, topic_set)
        self.wsapi_type = wsapi_type
        self.market_type = exchange.market_type

    def on_connected(self):
        if self.market_type != 'spot' and self.wsapi_type == 'private':
            self.pre_processors.append(self.on_error_message)

    def on_error_message(self, msg):
        if msg['op'] == 'close':
            raise RuntimeError('server closed')
        elif msg['op'] == 'error':
            raise RuntimeError('invalid op or inner error')
        else:
            return msg

    def create_keepalive_task(self):
        self.keepalive_msq = Queue()
        return super().create_keepalive_task()

    async def keepalive(self):
        while True:
            msg = await self.keepalive_msq.get()
            try:
                while True:
                    msg = self.keepalive_msq.get_nowait()
            except asyncio.QueueEmpty:
                pass

            if 'ping' in msg:
                # {"ping": 18212558000}
                pong = {'pong': msg['ping']}
            elif msg.get('op') == 'ping':
                # {"op": "ping", "ts": 1492420473058}
                pong = {'op': 'pong', 'ts': msg['ts']}
            else:
                # {"action": "ping", "data": {"ts": 1575537778295}}
                pong = {
                    'action': 'pong',
                    'data': {'ts': msg['data']['ts']}
                }
            await self.send(pong)

    def on_keepalive_message(self, msg):
        if ('ping' in msg or msg.get('op') == 'ping'
                or msg.get('action') == 'ping'):
            self.keepalive_msq.put_nowait(msg)
            raise StopIteration
        else:
            return msg

    @property
    def login_required(self):
        return 'private' in self.wsapi_type

    def on_login_message(self, msg):
        login_msg = False
        login_ok = False

        if msg.get('op') == 'auth':
            login_msg = True
            login_ok = (msg['err-code'] == 0)
        elif msg.get('action') == 'req' and msg.get('ch') == 'auth':
            login_msg = True
            login_ok = (msg['code'] == 200)

        if login_msg:
            if login_ok:
                self.logger.info(f'logged in')
                self.on_logged_in()
                raise StopIteration
            else:
                raise RuntimeError(f'login failed: {msg}')
        return msg

    def login_command(self, credentials):
        signature_method = 'HmacSHA256'
        apikey = credentials['apiKey']
        now = datetime.datetime.utcnow()
        timestamp = now.isoformat(timespec='seconds')

        if self.market_type == 'spot':
            params = keysort({
                'signatureMethod': signature_method,
                'signatureVersion': '2.1',
                'accessKey': apikey,
                'timestamp': timestamp
            })
        else:
            params = keysort({
                'SignatureMethod': signature_method,
                'SignatureVersion': '2',
                'AccessKeyId': apikey,
                'Timestamp': timestamp
            })

        auth = urllib.parse.urlencode(params)
        url = yarl.URL(self.wsurl)
        payload = '\n'.join(['GET', url.host, url.path, auth])
        signature_bytes = hmac(
            bytes(credentials['secret'], 'utf8'),
            bytes(payload, 'utf8'),
        )
        signature = signature_bytes.decode()

        if self.market_type == 'spot':
            params.update({
                'authType': 'api',
                'signature': signature
            })
            request = {
                'action': 'req',
                'ch': 'auth',
                'params': params
            }
        else:
            request = extend({
                'op': 'auth',
                'Signature': signature,
                'type': 'api'
            }, params)
        return request

    def create_subscribe_task(self):
        topics = {}
        for topic in self.topic_set:
            converted = self.convert_topic(topic)
            ch, params = self._split_params(converted)
            topics[ch] = params
        self.pre_processors.append(self.on_subscribe_message)
        self.pending_topics = set(topics)
        return self.awaitables.create_task(
            self.subscribe(topics), 'subscribe')

    def on_subscribe_message(self, msg):
        sub_msg = False
        sub_ok = False
        topic = None

        if 'subbed' in msg:  # huobipro & huobidm market
            sub_msg = True
            sub_ok = (msg['status'] == 'ok')
            topic = msg['subbed']
        elif msg.get('op') == 'sub':  # huobidm private
            sub_msg = True
            sub_ok = (msg['err-code'] == 0)
            topic = msg['topic']
        elif msg.get('action') == 'sub':  # huobipro private
            sub_msg = True
            sub_ok = (msg['code'] == 200)
            topic = msg['ch']

        if sub_msg:
            if sub_ok:
                self.logger.info(f'{topic} subscribed')
                self.on_subscribed(topic)
                raise StopIteration
            else:
                raise RuntimeError(f'subscribe failed: {msg}')
        return msg

    def subscribe_commands(self, topics):
        commands = []
        for ch, params in topics.items():
            if self.wsapi_type in ('private', 'public'):
                if self.market_type == 'spot':
                    request = {'action': 'sub', 'ch': ch}
                else:
                    request = {'op': 'sub', 'topic': ch}
            else:
                request = {'sub': ch}
            request.update(params)
            commands.append(request)
        return commands

    @staticmethod
    def _split_params(topic):
        ch, *params_string = topic.split('?', maxsplit=1)
        params = parse_qs(params_string[0]) if params_string else {}
        params = {k: lst[0] for k, lst in params.items()}
        return ch, params

    def decode(self, data):
        # huobipro private return str not bytes
        if isinstance(data, bytes):
            msg = gzip.decompress(data).decode()
        else:
            msg = data
        return json.loads(msg)


class HuobiWSReq(HuobiWSHandler):
    def __init__(self, exchange, wsapi_type):
        wsurl = exchange.urls['wsapi'][wsapi_type]
        super().__init__(exchange, wsurl, None, wsapi_type)
        self.queue = Queue()
        self.future = None
        self.timeout = 10.0  # in seconds

    def on_prepared(self):
        self.awaitables.create_task(self.sendreq(), 'sendreq')

    async def do_run(self, collector):
        await super().do_run(lambda r: self.future.set_result(r))

    async def sendreq(self):
        while True:
            self.future, req = await self.queue.get()
            await self.send(req)
            await asyncio.wait_for(self.future, self.timeout)

    def request(self, req):
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        self.queue.put_nowait((future, req))
        return future



class listiter:
    def __init__(self, lst):
        self.lst = lst
        self.cursor = 0 if lst else -1

    def has_next(self):
        return 0 <= self.cursor < len(self.lst)

    def __iter__(self):
        return self

    def __next__(self):
        if self.has_next():
            self.cursor += 1
            return self.lst[self.cursor - 1]
        raise StopIteration

    def rewind(self):
        self.cursor = 0 if self.lst else -1

    def prepend(self, elem):
        self.lst.insert(0, elem)
        self.cursor += 1

    def append(self, elem):
        self.lst.append(elem)
        if self.cursor < 0:
            self.cursor = 0

    def add(self, elem):
        i = max(self.cursor, 0)
        self.lst.insert(i, elem)
        self.cursor += 1

    def remove(self, elem=None):
        if elem is None:
            i = self.cursor - 1
        else:
            i = self.lst.index(elem)

        if i < 0:
            raise IndexError('no previous element')

        self.lst.pop(i)
        if self.cursor > 0:
            if i < self.cursor:
                self.cursor -= 1
        else:
            if not self.lst:
                self.cursor = -1

    def set(self, new_elem):
        if self.cursor > 0:
            self.lst[self.cursor - 1] = new_elem
        raise IndexError('no previous element')
